"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from abc import ABCMeta, abstractmethod
from collections.abc import Sized
from itertools import chain, product
from typing import Any, Dict, Iterable, List, Literal, Optional, Sequence, Set

import networkx as nx
from networkx.algorithms.isomorphism import DiGraphMatcher
from onnx import AttributeProto, NodeProto
from onnx.helper import get_attribute_value, make_attribute

from ..graph import OnnxGraph
from .utils import attribute_value


class Pattern(metaclass=ABCMeta):
    """Pattern to be matched. A pattern can be ORed and ADDed.

    Example::

        p1 = SingleNodePattern("Conv")
        p2 = SingleNodePattern("Relu")
        conv_or_relu = p1 | p2  # same as `p1 + p2`
    """

    @abstractmethod
    def match(
        self,
        graph: OnnxGraph,
        specify_node_names: Set[str] | None = None,
    ) -> Iterable[NodeProto | List[NodeProto]]:
        """Implementation how to match in the graph

        Args:
            graph (OnnxGraph): onnx graph
            specify_node_names (set[str], optional): Only optimize nodes with these
                names. Defaults to None.
        """

    def __or__(self, pattern: "Pattern") -> "Pattern":
        return OrPattern(self, pattern)

    def __add__(self, pattern: "Pattern") -> "Pattern":
        return OrPattern(self, pattern)

    def __radd__(self, pattern) -> "Pattern":
        """To support builtin functions like `sum`"""
        if isinstance(pattern, int):
            return self
        return OrPattern(self, pattern)


class OrPattern(Pattern):
    """A special pattern to match either p1 or p2."""

    def __init__(self, p1: Pattern, p2: Pattern):
        self.patterns: List[Pattern] = []
        if isinstance(p1, OrPattern):
            self.patterns.extend(p1.patterns)
        else:
            self.patterns.append(p1)
        if isinstance(p2, OrPattern):
            self.patterns.extend(p2.patterns)
        else:
            self.patterns.append(p2)

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        yield from chain(*(p.match(graph, specify_node_names) for p in self.patterns))


class SingleNodePattern(Pattern):
    """Match a single node type."""

    __id__ = 0

    def __init__(self, op_type: Optional[str] = None, op_name: Optional[str] = None):
        self.op_type = op_type
        self.op_name = op_name
        self.attr: List[AttributeProto] = []
        self.inputs: Optional[Sequence[str | None]] = None
        self.outputs: Optional[Sequence[str | None]] = None
        self.domain: Optional[str] = None
        self.order: Optional[Literal["pre", "post"]] = None
        self.id = SingleNodePattern.__id__
        SingleNodePattern.__id__ += 1

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        if isinstance(graph, OnnxGraph):
            if self.order == "pre":
                node_gen = nx.topological_sort(graph)
            elif self.order == "post":
                node_gen = nx.topological_sort(graph.reverse(copy=False))
            else:
                node_gen = iter(graph)
            for node in node_gen:
                node_pb = graph.nodes[node]["pb"]
                if self._check(node_pb, graph, specify_node_names):
                    yield node_pb
        else:  # match a single node
            node_pb = graph
            if self._check(node_pb, None, specify_node_names):
                yield node_pb

    def _check(self, node, graph, node_names):
        conditions = [
            self._check_type(node, self.op_type),
            self._check_attr(node),
            self._check_inputs(node, graph),
            self._check_outputs(node, graph),
            self._check_domain(node),
        ]
        if self.op_name:
            conditions.append(self._check_name(node, {self.op_name}))
        elif node_names:
            conditions.append(self._check_name(node, node_names))
        return all(conditions)

    def _check_type(self, node: NodeProto, op_type: str | None):
        return op_type == node.op_type or op_type is None

    def _check_name(self, node: NodeProto, names: Set[str]):
        return not names or node.name in names

    def _check_attr(self, node: NodeProto):
        if not self.attr:
            return True
        match_table: Dict[str, AttributeProto] = {}  # all attribute to check
        for attr in self.attr:
            match_table[attr.name] = attr
        matched = 0  # matched number of attributes
        for attr in node.attribute:
            if attr.name in match_table:
                matched += 1 if self._match_attr(match_table[attr.name], attr) else 0
        return matched == len(match_table)

    def _match_attr(self, attr0: AttributeProto, attr1: AttributeProto):
        def _is_any(v):
            if v.type == AttributeProto.STRING:
                return get_attribute_value(v).decode() == "__ANY__"
            return False

        if _is_any(attr0) or _is_any(attr1):
            return True
        if attr0.type != attr1.type:
            return False
        return attribute_value(attr0) == attribute_value(attr1)

    def _check_inputs(self, node: NodeProto, graph: OnnxGraph):
        if self.inputs is None or graph is None:
            return True
        input_nodes = [i.name for i in graph.onnx_predecessors(node)]
        for i in filter(lambda i: i.name in node.input, graph.initializer):
            input_nodes.append(i.name)
        if len(input_nodes) != len(self.inputs):
            return False
        return all(i is None or i in set(input_nodes) for i in self.inputs)

    def _check_outputs(self, node: NodeProto, graph: OnnxGraph):
        if self.outputs is None or graph is None:
            return True
        output_nodes = [i.name for i in graph.onnx_successors(node)]
        if len(output_nodes) != len(self.outputs):
            return False
        return all(i is None or i in set(output_nodes) for i in self.outputs)

    def _check_domain(self, node: NodeProto):
        if not self.domain:
            # do not compare if self.domain is not specified
            return self.domain is None or node.domain in ("", "ai.onnx")
        if "*" in self.domain or "?" in self.domain:
            # do not match ai.onnx
            if node.domain in ("", "ai.onnx"):
                return False
            pattern = re.compile(self.domain.replace("*", ".*").replace("?", "."))
            if re.findall(pattern, node.domain):
                return True
            return False
        return self.domain == node.domain

    def __hash__(self):
        op_type = self.op_type or ""
        op_name = self.op_name or ""
        return hash(f"{op_type}{op_name}{self.id}")

    def with_attr(
        self, name: str | AttributeProto, value: Optional[Any] = None
    ) -> "SingleNodePattern":
        """Match the node with specific attribute."""
        if isinstance(name, AttributeProto):
            self.attr.append(name)
        elif value is not None:
            self.attr.append(make_attribute(name, value))
        else:
            self.attr.append(make_attribute(name, "__ANY__"))
        return self

    def with_name(self, name: str) -> "SingleNodePattern":
        """Match the node with specific name."""
        assert isinstance(name, str)
        self.op_name = name
        return self

    def with_inputs(self, num_inputs: int, *input_nodes: str) -> "SingleNodePattern":
        """Match the node with specific number of inputs and names.

        Example::

            # a concat with 2 inputs
            SingleNodePattern("Concat").with_inputs(2)

            # an add with an input from node "X"
            SingleNodePattern("Add").with_inputs(2, "X", None)

        Args:
            num_inputs (int): number of inputs
            *input_nodes (str): names of each input node, this is not order sensitive,
                you can use `None` to match any input.
        """
        if input_nodes:
            assert len(input_nodes) == num_inputs
            self.inputs = list(input_nodes)
        else:
            self.inputs = [None] * num_inputs
        return self

    def with_outputs(self, num_outputs: int, *output_nodes: str) -> "SingleNodePattern":
        """Match the node with specific number of outputs and names.

        Example::

            # a split with 2 outputs
            SingleNodePattern("Split").with_outputs(2)

        Args:
            num_outputs (int): number of outputs
            *output_nodes (str): names of each output node, this is not order sensitive,
                you can use `None` to match any output.
        """
        if output_nodes:
            assert len(output_nodes) == num_outputs
            self.outputs = list(output_nodes)
        else:
            self.outputs = [None] * num_outputs
        return self

    def with_domain(self, domain: str) -> "SingleNodePattern":
        """Match the node with specific domain.

        Example::

            # a node with custom domain "onnxifier"
            SingleNodePattern().with_domain("onnxifier")

        Args:
            domain (str): domain name, if domain name is a wildcard
                (supports * and ? for now), match any domain using
                regex rules.
        """
        self.domain = str(domain)
        return self

    def with_order(self, order: Literal["pre", "post"]) -> "SingleNodePattern":
        """Match the node in specific order.

        Args:
            order (str):
            - "pre" to match the node in topological order.
            - "post" to match the node in reverse topological order.
        """
        self.order = order
        return self


class GraphPattern(Pattern, nx.DiGraph):
    """Match a subgraph."""

    def __init__(self, dag: Optional[nx.DiGraph] = None):
        if dag is not None:
            assert isinstance(dag, nx.DiGraph)
            assert all(isinstance(n, SingleNodePattern) for n in dag.nodes)
        super().__init__(dag)

    def add_node(self, node_for_adding, **attr) -> "GraphPattern":  # type: ignore
        assert isinstance(node_for_adding, Pattern)
        attr.update(pattern=node_for_adding)
        super().add_node(node_for_adding, **attr)
        return self

    def add_edge(self, u_of_edge, v_of_edge, **attr) -> "GraphPattern":  # type: ignore
        assert isinstance(u_of_edge, Pattern)
        assert isinstance(v_of_edge, Pattern)
        self.add_node(u_of_edge)
        self.add_node(v_of_edge)
        super().add_edge(u_of_edge, v_of_edge, **attr)
        return self

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        # Matching subgraph using nx method
        def _nm(g, h):
            return list(h["pattern"].match(g["pb"]))

        matcher = DiGraphMatcher(graph, self, node_match=_nm)
        for matches in matcher.subgraph_isomorphisms_iter():
            # match specify_node_names by "OR" pattern
            if specify_node_names:
                if any(key in specify_node_names for key in matches.keys()):
                    yield [graph.nodes[key]["pb"] for key in matches.keys()]
            else:
                yield [graph.nodes[key]["pb"] for key in matches.keys()]


class ConstantGraphPattern(Pattern):
    """Match each subgraph contains constant nodes"""

    def _is_deterministic(self, node) -> bool:
        return node.op_type not in {
            "RandomUniform",
            "RandomNormal",
            "RandomUniformLike",
            "RandomNormalLike",
            "Multinomial",
        }

    def _is_qdq(self, node) -> bool:
        return node.op_type in {
            "DequantizeLinear",
            "DynamicQuantizeLinear",
            "QuantizeLinear",
        }

    def _has_subgraph(self, node) -> bool:
        return any(
            attr.type in (AttributeProto.GRAPH, AttributeProto.GRAPHS)
            for attr in node.attribute
        )

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        const_names = {None, ""}  # for inputs with empty (dangled) name
        # get from initializers
        const_names.update(i.name for i in graph.initializer)
        const_nodes = []
        for _, node_name in enumerate(nx.topological_sort(graph)):
            node = graph.nodes[node_name]["pb"]
            if not self._is_deterministic(node) or self._is_qdq(node):
                continue
            if self._has_subgraph(node):
                continue
            if all(i in const_names for i in node.input):
                const_nodes.append(node)
                const_names.update(node.output)
        if specify_node_names:
            const_nodes = [n for n in const_nodes if n.name in specify_node_names]
        # TODO: cluster into subgraphs
        yield const_nodes


class OutputNodePattern(Pattern):
    """Match output nodes

    Args:
        match_all (bool): whether to match all output nodes or just one of them.
    """

    def __init__(self, match_all=False):
        self.match_all = match_all

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        nodes = []
        for output in graph.output:
            for node_name in reversed(list(nx.topological_sort(graph))):
                node = graph.nodes[node_name]["pb"]
                if output.name in node.output:
                    if self.match_all:
                        nodes.append(node)
                    elif not specify_node_names or node.name in specify_node_names:
                        yield node
        if self.match_all:
            if specify_node_names:
                nodes = [n for n in nodes if n.name in specify_node_names]
            yield nodes


class InputNodePattern(Pattern):
    """Match input nodes

    Args:
        match_all (bool): whether to match all output nodes or just one of them.
    """

    def __init__(self, match_all=False):
        self.match_all = match_all

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        nodes = []
        for input_var in graph.input:
            for node_name in nx.topological_sort(graph):
                node = graph.nodes[node_name]["pb"]
                if input_var.name in node.input:
                    if self.match_all:
                        nodes.append(node)
                    elif not specify_node_names or node.name in specify_node_names:
                        yield node
        if self.match_all:
            if specify_node_names:
                nodes = [n for n in nodes if n.name in specify_node_names]
            yield nodes


class StartEndPointPattern(Pattern):
    """Match a start-relay-end point pattern."""

    def __init__(self, start_pattern: Pattern, end_pattern: Pattern):
        self.start = start_pattern
        self.end = end_pattern

    def match(self, graph: OnnxGraph, specify_node_names: Set[str] | None = None):
        for beg, end in product(
            self.start.match(graph, specify_node_names),
            self.end.match(graph, specify_node_names),
        ):
            g = graph.copy(as_view=False)
            assert isinstance(g, OnnxGraph)
            if isinstance(beg, Sized) and len(beg) > 1:
                beg_node = self._merge_subgraph(g, beg)
            else:
                beg = beg[0] if isinstance(beg, Sequence) else beg
                beg_node = beg.name
            if isinstance(end, Sized) and len(end) > 1:
                end_node = self._merge_subgraph(g, end)
            else:
                end = end[0] if isinstance(end, Sequence) else end
                end_node = end.name
            # search for a path from start to end
            matched_nodes: Set[NodeProto] = set()
            try:
                paths = nx.shortest_simple_paths(g, beg_node, end_node)
                for i in [node for nodes in paths for node in nodes]:
                    if isinstance(i, nx.DiGraph):
                        matched_nodes.update(i.nodes[j]["pb"] for j in i)
                    else:
                        matched_nodes.add(graph.nodes[i]["pb"])
            except nx.NodeNotFound:
                continue
            yield list(matched_nodes)

    def _merge_subgraph(self, graph: nx.DiGraph, nodes):
        """Merge subgraph into a single node."""
        h: nx.DiGraph = graph.subgraph([n.name for n in nodes]).copy(  # type: ignore
            as_view=False,
        )
        pred_nodes: Set[NodeProto] = set()
        for in_node in filter(lambda i: h.in_degree(i) == 0, h):
            pred_nodes.update(graph.predecessors(in_node))
        succ_nodes: Set[NodeProto] = set()
        for out_node in filter(lambda o: h.out_degree(o) == 0, h):
            succ_nodes.update(graph.successors(out_node))
        graph.remove_nodes_from([n.name for n in nodes])
        for i in pred_nodes:
            graph.add_edge(i, h)
        for i in succ_nodes:
            graph.add_edge(h, i)
        return h



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Collection, List

import numpy as np
import onnx
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("fuse_qconv", deps=["infer_shape"])
class FuseQConvRewriter(Rewriter):
    """Fuse Quantize-Dequantize to QLinearConv.

    Before:

        DequantizeLinear -----
                              | (weight)
        DequantizeLinear -> Conv -> QuantizeLinear
                             /
        DequantizeLinear ___/  (bias, optional)

    After:

        QLinearConv

    Note:

        Bias typestr in QLinearConv must only be int32,
    """

    def __init__(self):
        qconv = GraphPattern()
        conv = SingleNodePattern("Conv")
        qconv.add_edge(SingleNodePattern("DequantizeLinear"), conv)
        qconv.add_edge(SingleNodePattern("DequantizeLinear"), conv)
        qconv.add_edge(conv, SingleNodePattern("QuantizeLinear"))
        super().__init__(qconv)

    def _unpack_nodes(self, nodes: Collection[NodeProto]):
        nodes = set(nodes)
        conv = next(filter(lambda n: n.op_type == "Conv", nodes))
        qinput = self.get_input_node(conv, 0)
        qweight = self.get_input_node(conv, 1)
        qoutput = None
        for n in self.get_output_node(conv, 0):
            if n.op_type == "QuantizeLinear":
                qoutput = n
        assert qinput in nodes and qinput is not None
        assert qweight in nodes and qweight is not None
        assert qoutput in nodes and qoutput is not None
        return qinput, qweight, conv, qoutput

    def _conv_attrs(self, conv):
        keys = ("auto_pad", "dilations", "group", "kernel_shape", "pads", "strides")
        attrs = {k: self.get_attribute(conv, k) for k in keys}
        return {k: v for k, v in attrs.items() if v is not None}

    def _canonical_bias(self, graph: OnnxGraph, conv, x_scale, w_scale):
        if (bias_value := self.get_value(conv.input[2])) is not None:
            # bias is a constant float
            # quantize bias to int32 implicitly
            scale = self.get_value_or_die(x_scale) * self.get_value_or_die(w_scale)
            qbias_value = np.round(bias_value / scale).astype(np.int32)
            graph.initializer.append(
                onnx.numpy_helper.from_array(qbias_value, conv.name + "/qbias")
            )
            return conv.name + "/qbias"
        if (bias := self.get_input_node_or_die(conv, 2)).op_type == "DequantizeLinear":
            # bias is already been quantized
            bias_value = self.get_value(bias.input[0])
            if bias_value is None:
                raise RuntimeError(
                    f"bias value from DequantizeLinear({bias.name}) is None"
                )
            if bias_value.dtype == np.int32:
                return bias.input[0]
            raise TypeError(
                f"QLinearConv only supports int32 bias, got {bias_value.dtype}"
            )
        if graph.tensor_type(conv.input[2]) != onnx.TensorProto.INT32:
            raise TypeError("QLinearConv only supports int32 bias.")
        return conv.input[2]

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        qinput, qweight, conv, qoutput = self._unpack_nodes(nodes)
        if len(graph.onnx_successors(qinput)) > 1:
            # this means input dequantize fans out to multiple downstream nodes,
            # we can't fuse it to QLinearConv
            return
        x = qinput.input[0]
        x_scale = qinput.input[1]
        x_zero_point = qinput.input[2]
        w = qweight.input[0]
        w_scale = qweight.input[1]
        w_zero_point = qweight.input[2]
        y_scale = qoutput.input[1]
        y_zero_point = qoutput.input[2]
        inputs = [
            x,
            x_scale,
            x_zero_point,
            w,
            w_scale,
            w_zero_point,
            y_scale,
            y_zero_point,
        ]
        if len(conv.input) == 3:
            bias = self._canonical_bias(graph, conv, x_scale, w_scale)
            inputs.append(bias)
        qconv = make_node(
            "QLinearConv",
            inputs=inputs,
            outputs=[qoutput.output[0]],
            name=conv.name + "/qconv",
            **self._conv_attrs(conv),  # type: ignore
        )
        self += qconv
        self -= [qinput, qweight, conv, qoutput]


@PASSES.register("unfuse_qconv")
class UnfuseQConvRewriter(Rewriter):
    """Unfuse QLinearConv to Quantize-Dequantize.

    Before:

        QLinearConv

    After:

        DequantizeLinear -----
                              | (weight)
        DequantizeLinear -> Conv -> QuantizeLinear
                             /
        DequantizeLinear ___/  (bias, optional)
    """

    def __init__(self):
        super().__init__(SingleNodePattern("QLinearConv"))

    def _conv_attrs(self, conv):
        keys = ("auto_pad", "dilations", "group", "kernel_shape", "pads", "strides")
        attrs = {k: self.get_attribute(conv, k) for k in keys}
        return {k: v for k, v in attrs.items() if v is not None}

    def _canonical_bias(self, graph: OnnxGraph, qconv):
        x_scale_value = self.get_value(qconv.input[1])
        w_scale_value = self.get_value(qconv.input[4])
        if x_scale_value is None:
            raise RuntimeError(f"x_scale value from {qconv.name} is not constant")
        if w_scale_value is None:
            raise RuntimeError(f"w_scale value from {qconv.name} is not constant")
        bias_value = self.get_value_or_die(qconv.input[8])
        bias_value = bias_value.astype(np.float32) * (x_scale_value * w_scale_value)
        graph.initializer.append(
            onnx.numpy_helper.from_array(bias_value, qconv.name + "/dq_bias")
        )
        return qconv.name + "/dq_bias"

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        x = node.input[0]
        x_scale = node.input[1]
        x_zero_point = node.input[2]
        w = node.input[3]
        w_scale = node.input[4]
        w_zero_point = node.input[5]
        y_scale = node.input[6]
        y_zero_point = node.input[7]
        if len(node.input) == 9:
            bias = [self._canonical_bias(graph, node)]
        else:
            bias = []
        dqinput = make_node(
            "DequantizeLinear",
            inputs=[x, x_scale, x_zero_point],
            outputs=[x + "_dq"],
            name=node.name + "/dqinput",
            axis=1,
        )
        dqweight = make_node(
            "DequantizeLinear",
            inputs=[w, w_scale, w_zero_point],
            outputs=[w + "_dq"],
            name=node.name + "/dqweight",
            axis=0,
        )
        conv = make_node(
            "Conv",
            inputs=[x + "_dq", w + "_dq"] + bias,
            outputs=[node.output[0] + "_q"],
            name=node.name + "/conv",
            **self._conv_attrs(node),  # type: ignore
        )
        qoutput = make_node(
            "QuantizeLinear",
            inputs=[node.output[0] + "_q", y_scale, y_zero_point],
            outputs=[node.output[0]],
            name=node.name + "/qoutput",
            axis=1,
        )
        self += [conv, dqinput, dqweight, qoutput]
        self -= node


@PASSES.register("unfuse_qmatmul")
class UnfuseQMatMulRewriter(Rewriter):
    """Unfuse QLinearMatMul to Quantize-Dequantize.

    Before:

        QLinearMatMul

    After:

        DequantizeLinear -----
                              | (weight)
        DequantizeLinear -> MatMul -> QuantizeLinear
                             /
        DequantizeLinear ___/  (bias, optional)
    """

    def __init__(self):
        super().__init__(SingleNodePattern("QLinearMatMul"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        a = node.input[0]
        a_scale = node.input[1]
        a_zero_point = node.input[2]
        b = node.input[3]
        b_scale = node.input[4]
        b_zero_point = node.input[5]
        y_scale = node.input[6]
        y_zero_point = node.input[7]
        dq_a = make_node(
            "DequantizeLinear",
            inputs=[a, a_scale, a_zero_point],
            outputs=[a + "_dq"],
            name=node.name + "/dqa",
        )
        dq_b = make_node(
            "DequantizeLinear",
            inputs=[b, b_scale, b_zero_point],
            outputs=[b + "_dq"],
            name=node.name + "/dqb",
        )
        matmul = make_node(
            "MatMul",
            inputs=[a + "_dq", b + "_dq"],
            outputs=[node.output[0] + "_q"],
            name=node.name + "/matmul",
        )
        qoutput = make_node(
            "QuantizeLinear",
            inputs=[node.output[0] + "_q", y_scale, y_zero_point],
            outputs=[node.output[0]],
            name=node.name + "/qoutput",
        )
        self += [dq_a, dq_b, matmul, qoutput]
        self -= node



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...... import ONNXIFIER_OPSET
from .. import OnnxGraph, Rewriter, cast_in
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="prelu")
class PReLU(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/activation/prelu-1.html

    https://onnx.ai/onnx/operators/onnx__PRelu.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        input_type = graph.tensor_type(ori_node.input[0])
        slope_type = graph.tensor_type(ori_node.input[1])
        if input_type != slope_type:
            self += cast_in(ori_node, 1, input_type)
        return make_node(
            "PRelu", inputs=ori_node.input, outputs=ori_node.output, name=ori_node.name
        )


@OP_CONVERT.register(name="relu")
class ReLU(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/activation/relu-1.html

    https://onnx.ai/onnx/operators/onnx__Relu.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        return make_node(
            "Relu", inputs=ori_node.input, outputs=ori_node.output, name=ori_node.name
        )


@OP_CONVERT.register(name="gelu")
class Gelu(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/activation/gelu-7.html

    https://onnx.ai/onnx/operators/onnx__Gelu.html
    """

    def __init__(self):
        super().__init__()
        self.register_post_hook(self.post_hook)

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        approximate = self.get_attribute(ori_node, "approximation_mode")
        assert isinstance(approximate, str)
        approximate = approximate.lower()
        return make_node(
            "Gelu",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            approximate="tanh" if approximate == "tanh" else "none",
        )

    def post_hook(self, graph: OnnxGraph):
        """Convert Gelu to subgraph for opset below 20."""
        # pylint: disable=import-outside-toplevel
        from onnxifier.passes.version_converter.opset19.gelu import Gelu as hook

        if ONNXIFIER_OPSET.version < 20:
            rewriter: Rewriter = hook()  # type: ignore
            return rewriter(graph)
        return graph


@OP_CONVERT.register(name="selu")
class Selu(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/activation/selu-1.html

    https://onnx.ai/onnx/operators/onnx__Selu.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        alpha = float(self.get_value_or_die(ori_node.input[1]).squeeze())
        gamma = float(self.get_value_or_die(ori_node.input[2]).squeeze())
        return make_node(
            "Selu",
            inputs=[ori_node.input[0]],
            outputs=ori_node.output,
            name=ori_node.name,
            alpha=alpha,
            gamma=gamma,
        )



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import List

import numpy as np
from onnx import TensorProto
from onnx.onnx_pb import NodeProto

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register("canonicalize_uint8_weights")
class CanonicalizeUint8WeightsRewriter(Rewriter):
    """Some devices (E.g. SM8650) do not support asymmetric weight quantization.
    If weights are quantized to uint8, this pass convert it to symmetric quantization.
    """

    def __init__(self):
        conv = SingleNodePattern("Conv")
        conv |= SingleNodePattern("ConvTranspose")
        conv |= SingleNodePattern("MatMul")
        pattern = GraphPattern()
        pattern.add_edge(SingleNodePattern("DequantizeLinear"), conv)
        super().__init__(pattern=pattern)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        dq_node, conv_node = nodes
        if self.get_input_node_or_die(conv_node, 1) != dq_node:
            # just to match the dequantize of weights, not the activations
            return
        x_dtype = graph.tensor_type(dq_node.input[0])
        if x_dtype != TensorProto.UINT8:
            return

        x = self.get_value_or_die(dq_node.input[0])
        x_scale = self.get_value_or_die(dq_node.input[1])
        x_zp = self.get_value_or_die(dq_node.input[2])
        if conv_node.op_type == "ConvTranspose":
            channel_axis = 1
        else:
            channel_axis = 0
        expand_axis = np.arange(x.ndim).tolist()
        expand_axis.pop(channel_axis)
        x_scale = np.expand_dims(x_scale, axis=expand_axis)
        x_zp = np.expand_dims(x_zp, axis=expand_axis)
        fx = (x.astype(np.float32) - x_zp) * x_scale
        x_max = np.abs(fx)
        for i in expand_axis:
            x_max = np.max(x_max, axis=i, keepdims=True)
        x_max = np.squeeze(x_max)
        x_new_scale = (x_max / np.iinfo(np.int8).max).astype(np.float32)
        x_new_scale[x_new_scale == 0] = 1
        x_new_zp = np.zeros_like(x_zp).reshape([-1]).astype(np.int8)
        x_new = (
            (fx / np.expand_dims(x_new_scale, axis=expand_axis))
            .round()
            .clip(-128, 127)
            .astype(np.int8)
        )
        x_cst = make_constant(f"{dq_node.name}/x", x_new)
        scale_cst = make_constant(f"{dq_node.name}/x_scale", x_new_scale)
        zp_cst = make_constant(f"{dq_node.name}/x_zp", x_new_zp)

        dq_node.input[0] = x_cst.output[0]
        dq_node.input[1] = scale_cst.output[0]
        dq_node.input[2] = zp_cst.output[0]
        self += [x_cst, scale_cst, zp_cst]



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
import pytest
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_operatorsetid,
    make_tensor_value_info,
)
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager


def _make_graph():
    gn = make_node(
        "GroupNormalization", ["X", "scale", "bias"], ["Y"], name="GN", num_groups=2
    )
    reshape = make_node("Reshape", ["Y", "shape"], ["Z"], name="reshape")
    add = make_node("Add", ["Z", "add_value"], ["W"], name="add")

    graph = make_graph(
        [gn, reshape, add],
        "test_graph",
        [make_tensor_value_info("X", onnx.TensorProto.FLOAT, [1, 32, 16, 16])],
        [make_tensor_value_info("W", onnx.TensorProto.FLOAT, [1, 32, 256])],
        [
            from_array(np.array([1, 1], np.float32), "scale"),
            from_array(np.array([0, 0], np.float32), "bias"),
            from_array(np.array([1, 32, 256], np.int64), "shape"),
            from_array(np.zeros([256], np.float32), "add_value"),
        ],
    )
    # GroupNormalization is deprecated after opset 18
    model = make_model(
        graph,
        ir_version=9,
        opset_imports=[make_operatorsetid("", 21)],
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def test_infer_shape_after_gn():
    graph = _make_graph()
    pm = PassManager(["infer_shape"])
    graph = pm.optimize(graph, strict=True)
    assert graph.tensor_shape("Y") == [1, 32, 16, 16]
    assert graph.tensor_shape("Z") == [1, 32, 256]


def _make_graph_sequence():
    split_to = make_node(
        "SplitToSequence",
        ["a", "split"],
        ["seq"],
        name="split",
        keepdims=0,
        axis=1,
    )
    seq_at = make_node(
        "SequenceAt",
        ["seq", "pos"],
        ["seq_out"],
        name="at",
    )
    trans = make_node(
        "Transpose", ["seq_out"], ["seq_out_t"], name="trans", perm=[0, 1]
    )
    reshape = make_node("Reshape", ["seq_out_t", "shape"], ["result"], name="reshape")
    graph = make_graph(
        [split_to, seq_at, trans, reshape],
        "seq_graph",
        [make_tensor_value_info("a", onnx.TensorProto.FLOAT, [2, 3, 4])],
        [make_tensor_value_info("result", onnx.TensorProto.FLOAT, [2, 4])],
        [
            from_array(np.array(1, np.int64), "split"),
            from_array(np.array(0, np.int64), "pos"),
            from_array(np.array([2, 4], np.int64), "shape"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model)
    return model


def test_infer_shape_with_sequence_at():
    model = _make_graph_sequence()
    pm = PassManager(["infer_shape"])
    graph = pm.optimize(OnnxGraph(model), strict=True)

    assert graph.tensor_shape("seq") == [2, 4]
    assert graph.tensor_shape("result") == [2, 4]
    graph._keep_value_info = True
    with pytest.raises(onnx.shape_inference.InferenceError):
        # onnx bugs
        onnx.checker.check_model(graph.model, True)
    # clear the value info to pass the checker
    graph._keep_value_info = False
    graph._value_info_update.clear()
    onnx.checker.check_model(graph.model, True)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Optional

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from onnxifier import logger

from .. import OnnxGraph
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="transpose")
class Transpose(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/movement/transpose-1.html

    https://onnx.ai/onnx/operators/onnx__Transpose.html
    """

    def _guess_perm(
        self, src_shape: List[int | str], target_shape: List[int | str]
    ) -> Optional[List[int]]:
        assert len(src_shape) == len(target_shape)
        # do not modify the original list directly
        src_shape_copy = src_shape.copy()
        perm = []
        for i in target_shape:
            perm.append(src_shape_copy.index(i))
            src_shape_copy[perm[-1]] = -1
        return perm

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        perm = self.get_value(ori_node.input[1])
        guessed_perm = None
        src_shape, _ = graph.tensor_info(ori_node.input[0])
        target_shape, _ = graph.tensor_info(ori_node.output[0])
        if src_shape and target_shape:
            guessed_perm = self._guess_perm(src_shape, target_shape)
        if perm is None or all(p == 0 for p in perm):
            logger.warning(f"perm is empty, use the guessed perm={guessed_perm}")
            perm = guessed_perm
        if perm is None:
            raise RuntimeError(
                f"Do not support transpose ({ori_node.name}) with dynamic perm"
            )
        ori_node.input.pop(1)
        return make_node(
            "Transpose",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            perm=list(map(int, perm)),
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnxscript
import torch
from onnx.numpy_helper import from_array
from onnxscript.onnx_opset import opset17 as op
from onnxscript.values import Opset

try:  # torch>=2.9
    from torch.onnx._internal.torchscript_exporter.jit_utils import GraphContext
    from torch.onnx._internal.torchscript_exporter.registration import onnx_symbolic
except ImportError:
    from torch.onnx._internal.jit_utils import GraphContext
    from torch.onnx._internal.registration import onnx_symbolic

from . import IR_DOMAIN


@onnxscript.script(Opset(domain=IR_DOMAIN.domain, version=IR_DOMAIN.version))
def diag(x, diagonal: int):
    """Implementation of ``aten::diag`` using ONNX operators.

    Note:

        This implementation only supports 1D input
    """

    shape = op.Shape(x)
    idv = op.ConstantOfShape(shape, value=from_array(np.array([1], np.float32)))
    idv_x = op.CastLike(idv, x)
    idv1d_t = op.Unsqueeze(idv_x, axes=0)
    x_1d = op.Unsqueeze(x, axes=1)
    y = op.MatMul(x_1d, idv1d_t)
    mask = op.EyeLike(y, k=diagonal)
    y = op.Mul(y, mask)
    return y


@onnx_symbolic("aten::diag", 1, custom=True)
@torch.onnx.symbolic_helper.parse_args("v", "i")
def diag_symbolic(g: GraphContext, x: torch._C.Value, diagonal: int = 0):
    """Symbolic function for ``torch.diag``."""

    return g.onnxscript_op(diag, x, diagonal_i=diagonal).setTypeAs(x)



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import numpy as np
import onnx
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.numpy_helper import from_array, to_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager


def _build_test_graph1():
    conv0 = make_node("Conv", inputs=["a", "w0"], outputs=["c"], group=4, name="conv0")
    conv1 = make_node("Conv", inputs=["c", "w1"], outputs=["d"], group=1, name="conv1")
    graph = make_graph(
        [conv0, conv1],
        name="graph",
        inputs=[],
        outputs=[make_value_info("d", make_tensor_type_proto(1, [1, 8, 124, 123]))],
        initializer=[
            from_array(np.random.normal(size=[1, 4, 128, 127]).astype("float32"), "a"),
            from_array(np.random.normal(size=[4, 1, 3, 3]).astype("float32"), "w0"),
            from_array(np.random.normal(size=[8, 4, 3, 3]).astype("float32"), "w1"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def _build_test_graph2():
    shape = make_node("Shape", inputs=["a"], outputs=["b"], name="shape")
    add = make_node("Add", inputs=["b", "x"], outputs=["c"], name="add")
    graph = make_graph(
        [shape, add],
        name="graph",
        inputs=[make_value_info("a", make_tensor_type_proto(1, [1, 32]))],
        outputs=[
            make_value_info("c", make_tensor_type_proto(onnx.TensorProto.INT64, [2]))
        ],
        initializer=[
            from_array(np.array([32, 1]).astype("int64"), "x"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def _build_test_graph3():
    shape = make_node("Shape", inputs=["a"], outputs=["b"], name="shape")
    add = make_node("Add", inputs=["b", "x"], outputs=["c"], name="add")
    graph = make_graph(
        [shape, add],
        name="graph",
        inputs=[make_value_info("a", make_tensor_type_proto(1, ["N", 32]))],
        outputs=[
            make_value_info("c", make_tensor_type_proto(onnx.TensorProto.INT64, [2]))
        ],
        initializer=[
            from_array(np.array([32, 1]).astype("int64"), "x"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def _build_test_graph4():
    conv = make_node("Conv", inputs=["x", "w", "b"], outputs=["y"], name="QConv")
    dq_w = make_node(
        "DequantizeLinear",
        inputs=["weight", "w_scale", "w_zp"],
        outputs=["w"],
        name="QWeight",
        axis=0,
    )
    reshape = make_node(
        "Reshape", inputs=["y", "shape"], outputs=["out"], name="Reshape"
    )
    add = make_node("Add", inputs=["p", "q"], outputs=["shape"], name="Add")
    graph = make_graph(
        [conv, dq_w, add, reshape],
        name="graph",
        inputs=[make_value_info("x", make_tensor_type_proto(1, [1, 3, 64, 64]))],
        outputs=[make_value_info("out", make_tensor_type_proto(1, [2, 4 * 64 * 64]))],
        initializer=[
            from_array(np.ones([8, 3, 1, 1], "int8"), "weight"),
            from_array(np.ones([8], "float32"), "w_scale"),
            from_array(np.zeros([8], "int8"), "w_zp"),
            from_array(np.random.normal(0, 0.1, size=[8]).astype("float32"), "b"),
            from_array(np.array([1, 2 * 64 * 64]).astype("int64"), "p"),
            from_array(np.array([1, 2 * 64 * 64]).astype("int64"), "q"),
        ],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    return model


def test_fold_constant():
    graph = OnnxGraph(_build_test_graph1())
    passes = PassManager(["fold_constant"])
    folded_graph = passes.optimize(graph, True)
    assert len(folded_graph) == 1


def test_fold_shape():
    graph = OnnxGraph(_build_test_graph2())
    passes = PassManager(["fold_constant"])
    folded_graph = passes.optimize(graph, True)
    assert len(folded_graph) == 1
    for node in folded_graph.nodes.values():
        assert np.all(to_array(node["pb"].attribute[0].t) == [33, 33])


def test_fold_dynamic_shape():
    graph = OnnxGraph(_build_test_graph3())
    passes = PassManager(["fold_constant"])
    folded_graph = passes.optimize(graph, True)
    assert len(folded_graph) == 3


def test_fold_should_bypass_qdq():
    graph = OnnxGraph(_build_test_graph4())
    passes = PassManager(["fold_constant"])
    folded_graph = passes.optimize(graph, True)
    assert any(
        folded_graph.nodes[node]["pb"].op_type == "DequantizeLinear"
        for node in folded_graph
    )



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="strided_slice")
class StridedSlice(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/movement/strided-slice-1.html

    https://onnx.ai/onnx/operators/onnx__Slice.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        # move steps to input[4]
        if len(ori_node.input) > 4:
            ori_node.input[4] = ori_node.input[3]
        else:
            ori_node.input.append(ori_node.input[3])
        ori_node.input[3] = ""
        begin_mask = self.get_attribute(ori_node, "begin_mask")
        assert isinstance(begin_mask, str)
        end_mask = self.get_attribute(ori_node, "end_mask")
        assert isinstance(end_mask, str)
        begin_mask = list(map(int, begin_mask.split(",")))
        end_mask = list(map(int, end_mask.split(",")))
        shrink_axis_mask = self.get_attribute(ori_node, "shrink_axis_mask")

        if any(i != 0 for i in begin_mask + end_mask):
            # adjust begin and end
            data_shape = graph.tensor_shape(ori_node.input[0])
            begin_var = self.get_value_or_die(ori_node.input[1]).copy()
            end_var = self.get_value_or_die(ori_node.input[2]).copy()
            for i, x in enumerate(begin_mask):
                begin_var[i] = 0 if x != 0 else begin_var[i]
            for i, x in enumerate(end_mask):
                end_var[i] = data_shape[i] if x != 0 else end_var[i]
            starts = make_constant(f"{ori_node.name}/starts", begin_var)
            ends = make_constant(f"{ori_node.name}/ends", end_var)
            ori_node.input[1] = starts.output[0]
            ori_node.input[2] = ends.output[0]
            self += [starts, ends]
        if not shrink_axis_mask:
            shrink_axis_mask = []
        else:
            assert isinstance(shrink_axis_mask, str)
            shrink_axis_mask = list(map(int, shrink_axis_mask.split(",")))
        if any(i != 0 for i in shrink_axis_mask):
            # add a squeeze after slice
            (axes,) = np.nonzero(shrink_axis_mask)  # type: ignore
            axes_node = make_constant(
                f"{ori_node.name}/Squeeze/axes", np.array(axes, dtype=np.int64)
            )
            squeeze = make_node(
                "Squeeze",
                name=f"{ori_node.name}/Squeeze",
                inputs=[f"{ori_node.name}/Slice_output0", axes_node.output[0]],
                outputs=[ori_node.output[0]],
            )
            ori_node.output[0] = f"{ori_node.name}/Slice_output0"
            self += [squeeze, axes_node]

        return make_node(
            "Slice",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
        )



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import networkx as nx

from ... import logger
from ...graph import OnnxGraph
from .. import L2


@L2.register()
def eliminate_dead_nodes(graph: OnnxGraph):
    """Remove nodes who doesn't connect to any output"""

    dead_nodes = set()
    exit_nodes = set(
        [graph.nodes[i]["pb"].name for i in graph.nodes if graph.nodes[i]["has_output"]]
    )
    for i in nx.topological_sort(graph.reverse()):
        if not any(nx.has_path(graph, i, j) for j in exit_nodes):
            dead_nodes.add(graph.nodes[i]["pb"])
            logger.debug(f"Dead node found: {i}")
    for node in dead_nodes:
        graph.remove_onnx_node(node)
    return graph



"""An Open Neural Network Exchange (ONNX) Optimization and Transformation Tool.

Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__version__ = "2.0.1"

import os
from copy import deepcopy
from typing import Any, Dict, Literal, Optional, Sequence

import onnx
from onnx import ModelProto
from onnx.helper import make_operatorsetid

from .domain import IR_DOMAIN, detect_domain, openvino_xml_to_onnx_graph
from .graph import OnnxGraph
from .pass_manager import PassManager, print_pass_simple
from .passes.version_converter.downgrade import downgrade_op_version
from .passes.version_converter.upgrade import upgrade_op_version


def convert_graph(
    model: str | os.PathLike | ModelProto,
    passes: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    onnx_format: Optional[Literal["protobuf", "textproto", "json", "onnxtxt"]] = None,
    strict: bool = False,
    configs: Optional[Dict[str, Any]] = None,
    print_passes: bool = True,
    target_opset: Optional[int] = None,
    recursive: bool = False,
    specify_node_names: Optional[Sequence[str]] = None,
) -> OnnxGraph:
    """Convert an ONNX model to OnnxGraph

    Args:
        model (str | os.PathLike | ModelProto): path to the model or a loaded model.
        passes (Sequence[str], optional): Names of selected passes. Defaults to None.
        exclude (Sequence[str], optional): Names of excluded passes. Defaults to None.
        onnx_format (str, optional): The serialization format of model file.
        strict (bool, optional): Break if any pass goes wrong. Defaults to False.
        configs (dict, optional): Specify configuration for passes
        print_passes (bool, optional): Print the selected passes. Defaults to True.
        target_opset (int, optional): Target opset version for ONNX domain. Defaults
            to ``ONNXIFIER_OPSET.version``.
        recursive (bool, optional): Apply passes to functions recursively.
        specify_node_names (Sequence[str], optional): Specify node names to apply
            passes only on those nodes.

    Returns:
        OnnxGraph: converted graph
    """
    base_dir = None
    for opset in detect_domain(model):
        if opset.domain == IR_DOMAIN.domain and opset.version <= IR_DOMAIN.version:
            model = openvino_xml_to_onnx_graph(model)
    if isinstance(model, (str, os.PathLike)):
        base_dir = os.path.dirname(model)
        model = onnx.load_model(model, format=onnx_format, load_external_data=False)
    else:
        model = deepcopy(model)
    graph = OnnxGraph(model, base_dir=base_dir)
    if target_opset is None:
        # align models to opset v19 because all passes is designed under opset19
        graph = upgrade_op_version(graph, op_version=ONNXIFIER_OPSET.version)
    pm = PassManager(passes, exclude=exclude, configs=configs)
    if print_passes:
        print_pass_simple(pm)
    node_names = set(specify_node_names) if specify_node_names else None
    graph = pm.optimize(
        graph,
        strict=strict,
        recursive=recursive,
        specify_node_names=node_names,
    )
    if target_opset is not None:
        if target_opset < graph.opset_version:
            graph = downgrade_op_version(graph, target_opset)
        else:
            graph = upgrade_op_version(graph, target_opset)
    return graph


def convert(
    model: str | os.PathLike | ModelProto,
    passes: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    onnx_format: Optional[Literal["protobuf", "textproto", "json", "onnxtxt"]] = None,
    strict: bool = False,
    configs: Optional[Dict[str, Any]] = None,
    print_passes: bool = True,
    target_opset: Optional[int] = None,
    recursive: bool = False,
    specify_node_names: Optional[Sequence[str]] = None,
) -> ModelProto:
    """Convert an ONNX model with default or given passes

    Args:
        model (str | os.PathLike | ModelProto): path to the model or a loaded model.
        passes (Sequence[str], optional): Names of selected passes. Defaults to None.
        exclude (Sequence[str], optional): Names of excluded passes. Defaults to None.
        onnx_format (str, optional): The serialization format of model file.
        strict (bool, optional): Break if any pass goes wrong. Defaults to False.
        configs (dict, optional): Specify configuration for passes.
        print_passes (bool, optional): Print the selected passes. Defaults to True.
        target_opset (int, optional): Target opset version for ONNX domain. Defaults
            to ``ONNXIFIER_OPSET.version``.
        recursive (bool, optional): Apply passes to functions recursively.
        specify_node_names (Sequence[str], optional): Specify node names to apply
            passes only on those nodes.
    """

    graph = convert_graph(
        model=model,
        passes=passes,
        exclude=exclude,
        onnx_format=onnx_format,
        strict=strict,
        configs=configs,
        print_passes=print_passes,
        target_opset=target_opset,
        recursive=recursive,
        specify_node_names=specify_node_names,
    )
    return graph.model


__all__ = ["convert", "convert_graph", "PassManager", "OnnxGraph"]

# make NodeProto hashable using node name
onnx.NodeProto.__hash__ = lambda self: hash(self.name)  # type: ignore

ONNXIFIER_IR_VERSION = onnx.IR_VERSION_2023_5_5
"""Currently used IR version, since most runtime supports up to this version."""

ONNXIFIER_OPSET = make_operatorsetid("", 19)
"""Currently used opset version, since most runtime supports up to this version."""



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import List

from onnx import NodeProto, TensorProto
from onnx.helper import make_node

from ...graph import OnnxGraph
from .. import PASSES, Rewriter
from ..pattern import SingleNodePattern


@PASSES.register(name="castlike_to_cast", deps=["infer_shape"])
class CastLikeToCastRewriter(Rewriter):
    """Convert CastLike to Cast, if data type is known to be static."""

    def __init__(self):
        super().__init__(SingleNodePattern("CastLike"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        dtype = graph.tensor_type(node.input[1])
        if dtype == TensorProto.UNDEFINED:
            return

        cast = make_node(
            "Cast",
            inputs=[node.input[0]],
            outputs=node.output,
            to=dtype,
            name=node.name,
        )

        self += cast
        self -= node



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=line-too-long

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


class BaseReduce(BaseNodeConversion):
    """Generic implementation of reduce operations."""

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        keepdims = self.get_attribute(ori_node, "keep_dims")
        assert isinstance(keepdims, str)
        keepdims = keepdims.lower() == "true"
        # canonicalize axes to int64
        axes = self.get_value_or_die(ori_node.input[1]).astype("int64")
        if axes.ndim == 0:
            axes = axes[None]  # axes must be a 1D tensor
        axes_node = make_constant(f"{ori_node.name}/axes", axes)
        ori_node.input[1] = axes_node.output[0]
        self += axes_node
        return make_node(
            self.__class__.__name__,
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            keepdims=1 if keepdims else 0,
        )


@OP_CONVERT.register(name="reduce_l1")
class ReduceL1(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-l1-4.html

    https://onnx.ai/onnx/operators/onnx__ReduceL1.html
    """


@OP_CONVERT.register(name="reduce_l2")
class ReduceL2(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-l2-4.html

    https://onnx.ai/onnx/operators/onnx__ReduceL2.html
    """


@OP_CONVERT.register(name="reduce_max")
class ReduceMax(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-max-1.html

    https://onnx.ai/onnx/operators/onnx__ReduceMax.html
    """


@OP_CONVERT.register(name="reduce_mean")
class ReduceMean(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-mean-1.html

    https://onnx.ai/onnx/operators/onnx__ReduceMean.html
    """


@OP_CONVERT.register(name="reduce_min")
class ReduceMin(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-min-1.html

    https://onnx.ai/onnx/operators/onnx__ReduceMin.html
    """


@OP_CONVERT.register(name="reduce_prod")
class ReduceProd(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-prod-1.html

    https://onnx.ai/onnx/operators/onnx__ReduceProd.html
    """


@OP_CONVERT.register(name="reduce_sum")
class ReduceSum(BaseReduce):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/reduction/reduce-sum-1.html

    https://onnx.ai/onnx/operators/onnx__ReduceSum.html
    """



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Protocol

from onnx import NodeProto

from .graph import OnnxGraph


class RewriterInterface(Protocol):
    """Interface helper for rewriting ONNX graphs."""

    __NAME__: str
    """Specify the name of the rewriter pass. Defaults to class or function name."""

    __DEPS__: List[str]
    """Specify the dependencies of the rewriter pass."""

    __PATCHES__: List[str]
    """Specify the finalizers of the rewriter pass."""

    def __call__(self, graph: OnnxGraph, *args, **kwargs) -> OnnxGraph:
        """Rewriter is a callable that takes :class:`OnnxGraph` and returns a
        modified graph.
        """
        return graph

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        """Implement how to rewrite matched nodes in the graph

        Args:
            graph (OnnxGraph): an onnx graph
            nodes (List[NodeProto]): a list of matched nodes
        """

    @property
    def num_rewrites(self) -> int:
        """This property records how many patterns have been matched and rewritten"""
        return 0



"""
Copyright (C) 2025 The ONNX2ONNX Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from itertools import chain
from pathlib import Path
from typing import List

import onnx

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import OutputNodePattern
from ..rewriter import Rewriter
from ..utils import canonical_node_name


@PASSES.register("split_non_compute_outputs", deps=["infer_shape"])
class SplitNonComputeOutputsRewriter(Rewriter):
    """Find non compute intensive nodes adjacent to the outputs and split them out.

    Example::

        Before:
            Conv -> Sigmoid -> Concat -<
                     Conv   __/

        After:
            1. Conv -<
               Conv -<
            2. Sigmoid -<

    """

    _compute_op = ("Conv", "ConvTranspose", "DeformConv", "Gemm", "MatMul")
    _fusing_act_op = ("Relu", "LeakyRelu", "PRelu")

    def __init__(self):
        self.model_cache = (
            Path("~").expanduser() / f".cache/onnx2onnx/{self.__class__.__name__}"
        )
        self.model_cache.mkdir(exist_ok=True, parents=True)
        super().__init__(OutputNodePattern())

    def _can_be_removed(self, graph: OnnxGraph, node, nodes_removed):
        if node.op_type in chain(self._compute_op, self._fusing_act_op):
            return False
        # all downstream nodes have been removed
        return all(i in nodes_removed for i in graph.onnx_successors(node))

    def rewrite(self, graph: OnnxGraph, nodes: List[onnx.NodeProto], *args, **kwargs):
        node = nodes[0]
        if node.op_type in self._compute_op:
            logger.debug(f"Skip node {node.name} since it's {node.op_type}")
            return
        nodes_to_remove: List[onnx.NodeProto] = []
        nodes_to_visit = [node]
        while nodes_to_visit:
            node_to_remove = nodes_to_visit.pop(0)
            if self._can_be_removed(graph, node_to_remove, nodes_to_remove):
                nodes_to_remove.append(node_to_remove)
                nodes_to_visit.extend(graph.onnx_predecessors(node_to_remove))
        # save the deleted nodes
        model_save = self.model_cache / f"{canonical_node_name(node.name)}.onnx"
        onnx.save_model(graph.onnx_subgraph(nodes_to_remove).model, model_save)
        logger.debug(f"Split out {len(nodes_to_remove)} nodes to {model_save}")
        self -= nodes_to_remove
        # add new output ports
        for leaf_node in filter(
            lambda i: i not in nodes_to_remove,
            chain(*[graph.onnx_predecessors(node) for node in nodes_to_remove]),
        ):
            for output_name in leaf_node.output:
                logger.debug(f"Add output to {leaf_node.name}:{output_name}")
                graph.set_output(leaf_node, output_name)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import onnx
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="convert")
class Convert(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/type/convert-1.html

    https://onnx.ai/onnx/operators/onnx__Cast.html
    """

    def _to_dtype(self, destination_type):
        match destination_type:
            case "f64":
                return onnx.TensorProto.DOUBLE
            case "f32":
                return onnx.TensorProto.FLOAT
            case "f16":
                return onnx.TensorProto.FLOAT16
            case "i64":
                return onnx.TensorProto.INT64
            case "i32":
                return onnx.TensorProto.INT32
            case "i16":
                return onnx.TensorProto.INT16
            case "i8":
                return onnx.TensorProto.INT8
            case "u64":
                return onnx.TensorProto.UINT64
            case "u32":
                return onnx.TensorProto.UINT32
            case "u16":
                return onnx.TensorProto.UINT16
            case "u8":
                return onnx.TensorProto.UINT8
        return onnx.TensorProto.UNDEFINED

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        destination_type = self.get_attribute(ori_node, "destination_type")
        assert isinstance(destination_type, str) or destination_type is None
        return make_node(
            "Cast",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            to=self._to_dtype(destination_type),
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
from onnx import numpy_helper
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_graph1(bias=True):
    qconv = make_node(
        "QLinearConv",
        inputs=[
            "x",
            "x_scale",
            "x_zero_point",
            "w",
            "w_scale",
            "w_zero_point",
            "y_scale",
            "y_zero_point",
        ],
        outputs=["y"],
        name="qconv",
        kernel_shape=[3, 3],
        pads=[1, 1, 1, 1],
        strides=[1, 1],
    )
    if bias:
        qconv.input.append("b")
    graph = make_graph(
        [qconv],
        "test_qconv",
        inputs=[make_tensor_value_info("x", onnx.TensorProto.UINT8, [1, 3, 16, 16])],
        outputs=[make_tensor_value_info("y", onnx.TensorProto.UINT8, [1, 16, 16, 16])],
        initializer=[
            numpy_helper.from_array(np.array(1 / 255, np.float32), "x_scale"),
            numpy_helper.from_array(np.array(0, np.uint8), "x_zero_point"),
            # avoid overflow
            numpy_helper.from_array(
                np.random.randint(0, 255, size=[16, 3, 3, 3], dtype=np.uint8), "w"
            ),
            numpy_helper.from_array(np.array([1 / 2048] * 16, np.float32), "w_scale"),
            numpy_helper.from_array(np.zeros([16], np.uint8), "w_zero_point"),
            numpy_helper.from_array(np.array(1 / 255, np.float32), "y_scale"),
            numpy_helper.from_array(np.array(0, np.uint8), "y_zero_point"),
            # this may occur some overflow
            numpy_helper.from_array(
                np.random.randint(-128 * 255, 127 * 255, size=[16], dtype=np.int32), "b"
            ),
        ],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def _build_graph2():
    m, n, k = 4, 8, 32
    qmatmul = make_node(
        "QLinearMatMul",
        inputs=[
            "a",
            "a_scale",
            "a_zero_point",
            "b",
            "b_scale",
            "b_zero_point",
            "y_scale",
            "y_zero_point",
        ],
        outputs=["y"],
        name="qmatmul",
    )
    graph = make_graph(
        [qmatmul],
        "test_qmatmul",
        inputs=[
            make_tensor_value_info("a", onnx.TensorProto.UINT8, [m, n]),
            make_tensor_value_info("b", onnx.TensorProto.UINT8, [n, k]),
        ],
        outputs=[make_tensor_value_info("y", onnx.TensorProto.UINT8, [m, k])],
        initializer=[
            numpy_helper.from_array(np.array(1 / 255, np.float32), "a_scale"),
            numpy_helper.from_array(np.array(0, np.uint8), "a_zero_point"),
            numpy_helper.from_array(np.array(1 / 255, np.float32), "b_scale"),
            numpy_helper.from_array(np.array(0, np.uint8), "b_zero_point"),
            numpy_helper.from_array(np.array(1 / 255, np.float32), "y_scale"),
            numpy_helper.from_array(np.array(0, np.uint8), "y_zero_point"),
        ],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    onnx.checker.check_model(model)
    return OnnxGraph(model)


def test_qconv_to_qdq_conv():
    graph = _build_graph1()
    run1 = Evaluator(graph.model, "OnnxRuntime")
    pm = PassManager(["unfuse_qconv"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) > 1
    run2 = Evaluator(graph.model, "OnnxRuntime")

    x = np.random.randint(0, 255, size=[1, 3, 16, 16], dtype=np.uint8)
    y1 = run1(["y"], {"x": x})[0]
    y2 = run2(["y"], {"x": x})[0]
    err = np.abs(y1.astype(np.int32) - y2.astype(np.int32))
    assert err.max() <= 1

    pm = PassManager(["fuse_qconv"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 1
    run3 = Evaluator(graph.model, "OnnxRuntime")
    y3 = run3(["y"], {"x": x})[0]
    err = np.abs(y1.astype(np.int32) - y3.astype(np.int32))
    assert err.max() <= 1
    assert not (np.all(y1 == 0) or np.all(y1 == 255))


def test_qconv_to_qdq_conv_without_bias():
    graph = _build_graph1(bias=False)
    run1 = Evaluator(graph.model, "OnnxRuntime")
    pm = PassManager(["unfuse_qconv"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) > 1
    run2 = Evaluator(graph.model, "OnnxRuntime")

    x = np.random.randint(0, 255, size=[1, 3, 16, 16], dtype=np.uint8)
    y1 = run1(["y"], {"x": x})[0]
    y2 = run2(["y"], {"x": x})[0]
    err = np.abs(y1.astype(np.int32) - y2.astype(np.int32))
    assert err.max() <= 1

    pm = PassManager(["fuse_qconv"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 1
    run3 = Evaluator(graph.model, "OnnxRuntime")
    y3 = run3(["y"], {"x": x})[0]
    err = np.abs(y1.astype(np.int32) - y3.astype(np.int32))
    assert err.max() <= 1
    assert not (np.all(y1 == 0) or np.all(y1 == 255))


def test_qmatmul_to_qdq_matmul():
    graph = _build_graph2()
    run1 = Evaluator(graph.model, "OnnxRuntime")
    pm = PassManager(["unfuse_qmatmul"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) > 1
    run2 = Evaluator(graph.model, "OnnxRuntime")

    a = np.random.randint(0, 255, size=[4, 8], dtype=np.uint8)
    b = np.random.randint(0, 255, size=[8, 32], dtype=np.uint8)
    y1 = run1(["y"], {"a": a, "b": b})[0]
    y2 = run2(["y"], {"a": a, "b": b})[0]
    err = np.abs(y1.astype(np.int32) - y2.astype(np.int32))
    assert err.max() <= 1



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import itertools
from collections import defaultdict
from typing import Dict, List, Optional, Sequence, Set

import networkx as nx
import numpy as np
from onnx import NodeProto
from onnx.numpy_helper import to_array

from ... import logger
from ...graph import OnnxGraph
from .. import PASSES, Rewriter
from ..pattern import SingleNodePattern
from ..utils import evaluate_on_node, make_constant


def _ensure_simple_batch(node_name, indices: Set[int]):
    if len(indices) > 1:
        raise NotImplementedError(
            f"The input of operator {node_name} has complex batch definition."
        )


def _get_dim_after_transpose(node_pb: NodeProto, index: int) -> int:
    perm: List[int] = []
    for t in node_pb.attribute:
        if t.name == "perm":
            perm = list(t.ints)
    assert perm
    return perm.index(index)


def _get_dim_after_reshape(graph: OnnxGraph, node_pb: NodeProto, index: int) -> int:
    # To guess the dimension merge and split, it not supported on symbolic shape.
    input_shape = graph.static_tensor_shape(node_pb.input[0])
    output_shape = graph.static_tensor_shape(node_pb.output[0])
    batch = input_shape[index]
    in_s1, in_s2 = input_shape[:index], input_shape[index + 1 :]
    if batch in output_shape and output_shape.count(batch) == 1:
        out_ind = output_shape.index(batch)
        out_s1, out_s2 = output_shape[:out_ind], output_shape[out_ind + 1 :]
        if (np.prod(in_s1) == np.prod(out_s1)) and (np.prod(in_s2) == np.prod(out_s2)):
            return out_ind
        else:
            raise ValueError(
                f"Can't trace complex reshape {input_shape} -> {output_shape}"
            )
    # Search for (i, j, k) that index in [i, j) and
    # prod(input_shape[i:j]) == output_shape[k]
    for i, j in itertools.product(
        range(index - 1, -1, -1), range(index, len(input_shape))
    ):
        for k, _ in enumerate(output_shape):
            if np.prod(input_shape[i : j + 1]) == output_shape[k]:
                logger.debug(f"Found batch in output shape {output_shape}[{k}]")
                return k

    raise NotImplementedError(f"Complex reshape {input_shape} -> {output_shape}")


def _canonicalize_neg_axes(axes, rank: int) -> List[int]:
    if axes.ndim == 0:
        axes = axes[None]
    axes = list(axes)
    for i, axis in enumerate(axes):
        if axis < 0:
            axes[i] = rank + axis
    return axes


def _get_dim_after_squeeze(graph: OnnxGraph, node_pb: NodeProto, index: int) -> int:
    axes: Optional[Sequence[int] | np.ndarray] = []
    if node_pb.input[1] in graph.initializers:
        axes = to_array(graph.initializers[node_pb.input[1]])
    else:
        axes_node = graph.onnx_predecessors(node_pb)[-1]
        axes = evaluate_on_node(graph, axes_node)
    if axes is None:
        raise ValueError(f'Can\'t trace squeeze "{node_pb.name}" with unknown axes')
    rank = len(graph.tensor_shape(node_pb.input[0]))
    axes = _canonicalize_neg_axes(axes, rank)
    if index in axes:
        raise ValueError(f'Batch is squeezed out in "{node_pb.name}"')
    for i in sorted(axes):
        if i < index:
            index -= 1
    assert index >= 0
    return index


def _get_dim_after_unsqueeze(graph: OnnxGraph, node_pb: NodeProto, index: int) -> int:
    axes: Optional[Sequence[int] | np.ndarray] = []
    if node_pb.input[1] in graph.initializers:
        axes = to_array(graph.initializers[node_pb.input[1]])
    else:
        axes_node = graph.onnx_predecessors(node_pb)[-1]
        axes = evaluate_on_node(graph, axes_node)
    if axes is None:
        raise ValueError(f'Can\'t trace unsqueeze "{node_pb.name}" with unknown axes')
    rank = len(graph.tensor_shape(node_pb.input[0]))
    # note unsqueeze(x, -1) is equivalent to unsqueeze(x, x.ndim)
    axes = _canonicalize_neg_axes(axes, rank + 1)
    for i in sorted(axes):
        if i <= index:
            index += 1
    return index


def _get_dim_after_gather(graph: OnnxGraph, node_pb: NodeProto, index: int) -> int:
    axis = 0
    for t in node_pb.attribute:
        if t.name == "axis":
            axis = t.i
    rank = len(graph.tensor_shape(node_pb.input[1]))
    if axis == index and rank == 0:
        raise ValueError(f'Gather "{node_pb.name}" eliminates the batch dimension')
    return index


def _dispatch_shuffle_op(graph: OnnxGraph, node_pb: NodeProto, index: int) -> int:
    if node_pb.op_type == "Transpose":
        index = _get_dim_after_transpose(node_pb, index)
    elif node_pb.op_type == "Reshape":
        index = _get_dim_after_reshape(graph, node_pb, index)
    elif node_pb.op_type == "Squeeze":
        index = _get_dim_after_squeeze(graph, node_pb, index)
    elif node_pb.op_type == "Unsqueeze":
        index = _get_dim_after_unsqueeze(graph, node_pb, index)
    elif node_pb.op_type == "Gather":
        index = _get_dim_after_gather(graph, node_pb, index)
    return index


def trace_batch_dimension(
    graph: OnnxGraph, batch_index: Optional[Dict[str, int]] = None
):
    """trace the batch dimension of the graph.

    Args:
        graph (OnnxGraph): The input graph.
        batch_index (Dict[str, int]): The argument to specify the index of the batch
            dimension of each inputs. Defaults to all 0.
    """

    if batch_index is None:
        batch_index = {i: 0 for i in graph.inputs}
    for name, idx in batch_index.items():
        logger.debug(f'batch dim of "{name}" is {idx}')

    # tracking the dimension index of the inputs,
    # change only after {transpose, reshape}.
    input_index_tracing: Dict[str, Set[int]] = defaultdict(set)
    output_index_tracing: Dict[str, Set[int]] = defaultdict(set)
    for node_name in nx.topological_sort(graph):
        node_pb: NodeProto = graph.nodes[node_name]["pb"]
        if graph.nodes[node_name]["has_input"]:
            indices: Set[int] = set()
            for input_name in node_pb.input:
                if input_name in batch_index:
                    indices.add(batch_index[input_name])
            assert indices
            _ensure_simple_batch(node_name, indices)
            index = indices.pop()
            index = _dispatch_shuffle_op(graph, node_pb, index)
            output_index_tracing[node_name].add(index)
            for i in graph.successors(node_name):
                input_index_tracing[i].add(index)
        elif node_name in input_index_tracing:
            indices = input_index_tracing[node_name].copy()
            _ensure_simple_batch(node_name, indices)
            index = indices.pop()
            index = _dispatch_shuffle_op(graph, node_pb, index)
            output_index_tracing[node_name].add(index)
            for i in graph.successors(node_name):
                input_index_tracing[i].add(index)

    for name, indices in output_index_tracing.items():
        _ensure_simple_batch(name, indices)
        logger.trace(f"{name}: batch dim = {indices}")

    return output_index_tracing


@PASSES.register("reset_reshape_batch")
class ResetReshapeBatchRewriter(Rewriter):
    """Reset the batch dimension of Reshape to -1.

    Batch dimension is tracked by :func:`trace_batch_dimension` function.
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("Reshape"))
        self.index_tracing: Dict[str, Set[int]] = {}

    def match_and_rewrite(self, graph: OnnxGraph, *args, **kwargs) -> OnnxGraph:
        self.index_tracing = trace_batch_dimension(graph, kwargs.get("batch_index"))
        return super().match_and_rewrite(graph, *args, **kwargs)

    def rewrite(self, graph, nodes, *args, **kwargs):
        node = nodes[0]
        shape = self.get_value(node.input[1])
        if shape is None:
            logger.debug(f'target shape of "{node.name}" is dynamic')
            return

        if node.name in self.index_tracing:
            indices = self.index_tracing[node.name].copy()
            if len(indices) != 1:
                return
            index = indices.pop()
            # fill exist -1
            output_shape = graph.tensor_shape(node.output[0])
            shape = list(shape)
            for i, s in enumerate(shape):
                if s == -1:
                    shape[i] = output_shape[i]
            # reset batch dim to -1
            shape[index] = -1
            shape_cst = make_constant(f"{node.name}/shape", np.array(shape, np.int64))
            node.input[1] = shape_cst.output[0]
            self += shape_cst



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List, Sequence

import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(
    name="resize_move_size_to_scale",
    deps=["initializer_to_constant", "infer_shape", "shape_to_constant"],
)
class ResizeMoveSizeToScaleRewriter(Rewriter):
    """Move `size` input to `scale` input for Resize Op."""

    def __init__(self):
        super().__init__(SingleNodePattern("Resize"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node_pb = nodes[0]
        node = node_pb.name
        input_nodes = self.get_input_nodes(node_pb)
        if len(input_nodes) < 3:
            raise ValueError(
                f"Op '{node}' both scales and sizes are empty!"
                " Try fold_constant pass before this."
            )
        elif len(input_nodes) == 3:
            return
        _, roi, scales, sizes = input_nodes
        if sizes is None and scales is None:
            raise ValueError(
                f"Op '{node}' both scales and sizes are empty!"
                " Try fold_constant pass before this."
            )
        if scales is not None:
            return
        assert sizes is not None
        input_shape = graph.static_tensor_shape(node_pb.input[0])
        axes = self.get_attribute(node_pb, "axes") or range(len(input_shape))
        ct_mode = self.get_attribute(node_pb, "coordinate_transformation_mode")
        sizes_val = self.get_value_or_die(sizes)
        assert isinstance(axes, Sequence)
        axes = [int(i) for i in axes]  # type: ignore
        if roi is not None and ct_mode == "tf_crop_and_resize":
            roi_val = self.get_value_or_die(roi).reshape([2, -1])
            roi_size = []
            for i, j, k in zip(roi_val[0], roi_val[1], sizes_val):
                if i < 0:
                    i += k
                if j < 0:
                    j += k
                assert j >= i >= 0
                roi_size.append(j - i)
            scales_val = [sizes_val[i] / roi_size[i] for i, _ in enumerate(axes)]
        else:
            scales_val = [sizes_val[i] / input_shape[j] for i, j in enumerate(axes)]
        scales = make_constant(
            f"{node}/const/scales", np.array(scales_val, dtype="float32")
        )
        node_pb.input[2] = scales.output[0]
        node_pb.input.pop()  # remove `sizes`
        self += scales
        self -= sizes



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.onnx_pb import NodeProto

from . import IR_PASSES, OnnxGraph, Rewriter, SingleNodePattern, make_constant


@IR_PASSES.register("shapeof_to_constant")
class ShapeOfToConstantRewriter(Rewriter):
    """Calculate the result of ShapeOf node and replace it with a Constant node.

    If shape is dynamic, do nothing.
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("ShapeOf"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        input_shape, _ = graph.tensor_info(node.input[0])
        if input_shape is None:
            return
        elif any(isinstance(i, str) or i <= 0 for i in input_shape):
            return  # dynamic shape

        self -= node
        if self.get_attribute(node, "output_type") == "i32":
            dtype = "int32"
        else:
            dtype = "int64"
        cst_node = make_constant(node.name, np.array(input_shape, dtype))
        cst_node.output[0] = node.output[0]
        self += cst_node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx import NodeProto
from onnx.helper import make_node, make_tensor

from .... import OnnxGraph
from ...pattern import SingleNodePattern
from ...rewriter import Rewriter
from ...utils import make_constant
from . import OP_CONVERTER


@OP_CONVERTER.register("Col2Im")
class Col2Im(Rewriter):
    """Convert Col2Im v18 to equivalent counterpart using v17 supported operations.

    Col2Im rearranges column blocks back into a multidimensional image.
    Since Col2Im doesn't exist in opset 17, we implement it using supported operations:
    Reshape, Transpose, and other basic tensor operations.
    """

    def __init__(self):
        super().__init__(SingleNodePattern("Col2Im"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        dilations = self.get_attribute(node, "dilations")
        if isinstance(dilations, list) and any(i != 1 for i in dilations):
            raise ValueError("Dilation is not supported in Col2Im conversion.")

        # Get block_shape from input to determine spatial dimensions
        block_shape = self.get_value_or_die(node.input[2])
        spatial_rank = len(block_shape)

        stride = self.get_attribute(node, "strides")
        padding = self.get_attribute(node, "pads")
        if stride is None:
            stride = [1] * spatial_rank
        if padding is None:
            padding = [0] * (2 * spatial_rank)
        assert isinstance(stride, list) and isinstance(padding, list)
        stride = [int(s) for s in stride if isinstance(s, (int, float))]
        padding = [int(p) for p in padding if isinstance(p, (int, float))]

        # Try the simple case first: use DepthToSpace if possible
        if self._try_depth_to_space_conversion(graph, node, stride, padding):
            return

        # Fall back to general implementation using basic operations
        self._general_col2im_implementation(graph, node, stride, padding)

    def _try_depth_to_space_conversion(
        self,
        graph: OnnxGraph,
        node: NodeProto,
        stride: List[int],
        padding: List[int],
    ) -> bool:
        """Try to convert Col2Im to DepthToSpace for simple cases."""

        output_shape = graph.tensor_shape(node.output[0])[2:]
        if all(isinstance(i, int) for i in output_shape):
            image_shape = [int(i) for i in output_shape]
        else:
            image_shape = self.get_value_or_die(node.input[1])
        block_shape = self.get_value_or_die(node.input[2])
        if not all(block_shape == stride):
            return False
        if any(i != block_shape[0] for i in block_shape):
            return False
        if any(i != 0 for i in padding):
            return False
        input_shape = graph.tensor_shape(node.input[0])
        # Simple case: can use DepthToSpace
        b, c, _ = input_shape
        ds_shape = [b, c, *[i // s for i, s in zip(image_shape, stride)]]
        shape_cst = make_constant(
            f"{node.name}/shape", np.array(ds_shape, dtype=np.int64)
        )
        reshape = make_node(
            "Reshape",
            [node.input[0], shape_cst.output[0]],
            [f"{node.name}/Reshape_output0"],
            name=f"{node.name}/Reshape",
        )
        d2s = make_node(
            "DepthToSpace",
            [reshape.output[0]],
            [node.output[0]],
            name=node.name,
            blocksize=int(block_shape[0]),
            mode="CRD",
        )

        self += [shape_cst, reshape, d2s]
        self -= node
        return True

    def _general_col2im_implementation(
        self,
        graph: OnnxGraph,
        node: NodeProto,
        stride: List[int],
        padding: List[int],
    ):
        """General Col2Im implementation using ScatterElements for complex cases.

        Col2Im rearranges column data back to image format.
        Implementation: Col2Im(x) = Crop(ScatterElements(z, indices, x))
        where z is zero tensor with padded output shape.
        """

        # Get input dimensions and parameters
        input_shape = graph.static_tensor_shape(node.input[0])  # [B, C*kH*kW, L]
        block_shape = self.get_value_or_die(node.input[2])  # [kH, kW]
        image_shape = self.get_value_or_die(node.input[1])  # [H, W]
        dtype = graph.tensor_type(node.output[0])

        # Convert to integers, handling dynamic shapes if needed
        batch = int(input_shape[0])
        channels_times_kernel = int(input_shape[1])
        L = int(input_shape[2])

        kernel_size = int(np.prod(block_shape))
        channels = channels_times_kernel // kernel_size

        # Calculate padded dimensions
        paddings = np.reshape(padding, (2, -1)).T
        image_pad_shape = [i + p[0] + p[1] for i, p in zip(image_shape, paddings)]

        # Step 1: Reshape input from [B, C*kH*kW, L] to [B, C, kH*kW, L]
        update_shape = make_constant(
            f"{node.name}/update_shape",
            np.array([batch, channels, kernel_size * L], dtype=np.int64),
        )
        update_reshape = make_node(
            "Reshape",
            [node.input[0], update_shape.output[0]],
            [f"{node.name}/update_reshape_output0"],
            name=f"{node.name}/update_reshape",
        )

        # Step 2: Create zero tensor z with padded shape [B, C, padded_H*padded_W]
        flatten_shape = make_constant(
            f"{node.name}/flatten_shape",
            np.array([batch, channels, np.prod(image_pad_shape)], dtype=np.int64),
        )

        zero_tensor = make_node(
            "ConstantOfShape",
            [flatten_shape.output[0]],
            [f"{node.name}/zero_tensor"],
            name=f"{node.name}/zero_tensor",
            value=make_tensor("value", dtype, [1], [0]),
        )

        # Step 3: Calculate ScatterElements indices
        # We need to map each position of input block [kH, kW, L]
        # to the correct position in output tensor [H, W]
        # and it is repeated on batch size and channel.

        # Calculate how many patches fit in each dimension of the output
        spatial_shape = [
            int(np.ceil((i + p[0] + p[1] - k + 1) / s))
            for i, p, k, s in zip(image_shape, paddings, block_shape, stride)
        ]
        assert L == np.prod(spatial_shape)
        # Input X (omit batch and channel) has shape [*block_shape, *spatial_shape]
        block_indices = np.meshgrid(*[np.arange(d) for d in block_shape], indexing="ij")
        block_indices = np.stack(block_indices, axis=-1)
        block_indices = block_indices.reshape(-1, len(block_shape))
        spatial_indices = np.meshgrid(
            *[np.arange(d) for d in spatial_shape], indexing="ij"
        )
        spatial_indices = np.stack(spatial_indices, axis=-1)
        spatial_indices = spatial_indices.reshape(-1, len(spatial_shape))
        # Create all combinations of kernel and spatial indices
        kdx = np.repeat(block_indices, len(spatial_indices), axis=0)
        sdx = np.tile(spatial_indices, (len(block_indices), 1))
        # Calculate output positions using correct Col2Im formula
        output_positions = []
        for i in range(len(spatial_shape)):
            pos = sdx[:, i] * stride[i] + kdx[:, i]  # + paddings[i, 0]
            output_positions.append(pos)
        odx = np.stack(output_positions, axis=1)

        # Create indices array [kernel_size * L]
        indices_array = np.ravel_multi_index(odx.T, image_pad_shape)  # type: ignore
        indices_array = np.tile(indices_array, [batch, channels, 1])
        indices_array = indices_array.reshape([batch, channels, -1])

        indices_const = make_constant(
            f"{node.name}/indices",
            indices_array,
        )

        # Step 4: Apply ScatterElements to place patches
        # ScatterElements: scatter updates into zero tensor based on indices
        scatter = make_node(
            "ScatterElements",
            [zero_tensor.output[0], indices_const.output[0], update_reshape.output[0]],
            [f"{node.name}/scattered"],
            name=f"{node.name}/scatter",
            axis=2,  # Scatter along spatial dimension,
            reduction="add",
        )

        # Step 5: Reshape back to spatial dimensions [B, C, padded_H, padded_W]
        padded_shape = make_constant(
            f"{node.name}/padded_shape",
            np.array([batch, channels, *image_pad_shape], dtype=np.int64),
        )
        reshape_back = make_node(
            "Reshape",
            [scatter.output[0], padded_shape.output[0]],
            [f"{node.name}/padded_result"],
            name=f"{node.name}/reshape_back",
        )

        # Step 6: Crop padding if needed
        if any(p > 0 for p in padding):
            starts = make_constant(
                f"{node.name}/crop_starts",
                np.concatenate([[0, 0], paddings[:, 0]]).astype(np.int64),
            )
            image_size = [i + p for i, p in zip(image_shape, paddings[:, 0])]
            ends = make_constant(
                f"{node.name}/crop_ends",
                np.concatenate([[batch, channels], image_size]).astype(np.int64),
            )
            crop = make_node(
                "Slice",
                [reshape_back.output[0], starts.output[0], ends.output[0]],
                [node.output[0]],
                name=node.name,
            )

            nodes_to_add = [
                update_shape,
                update_reshape,
                flatten_shape,
                indices_const,
                padded_shape,
                starts,
                ends,
                zero_tensor,
                scatter,
                reshape_back,
                crop,
            ]
        else:
            # No padding needed, use result directly
            final_rename = make_node(
                "Identity",
                [reshape_back.output[0]],
                [node.output[0]],
                name=node.name,
            )

            nodes_to_add = [
                update_shape,
                update_reshape,
                flatten_shape,
                indices_const,
                padded_shape,
                zero_tensor,
                scatter,
                reshape_back,
                final_rename,
            ]

        self += nodes_to_add
        self -= node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

import itertools
import re

import onnx
from onnx import TensorProto
from onnx.helper import make_attribute
from onnx.numpy_helper import from_array, to_array

from ... import OnnxGraph, logger
from .. import L1, PASSES
from ..utils import cast_in


@L1.register()
def half_to_float(graph: OnnxGraph) -> OnnxGraph:
    """Convert half consts and values to float32."""

    for node in graph:
        node_pb = graph.nodes[node]["pb"]
        if node_pb.op_type == "Constant":
            tensor = node_pb.attribute[0].t
            if tensor.data_type == TensorProto.FLOAT16:
                array = to_array(tensor).astype("float32")
                attr = make_attribute(key="value", value=from_array(array))
                node_pb.attribute.pop()
                node_pb.attribute.append(attr)
        elif node_pb.op_type == "Cast":
            # if cast target is fp16, change it to fp32
            for attr in node_pb.attribute:
                if attr.name == "to" and attr.i == TensorProto.FLOAT16:
                    attr.i = TensorProto.FLOAT
    for init in graph.initializer:
        if init.data_type == TensorProto.FLOAT16:
            array = to_array(init).astype("float32")
            init.data_type = TensorProto.FLOAT
            init.raw_data = from_array(array).raw_data
    for io in itertools.chain(graph.input, graph.output):
        if io.type.tensor_type.elem_type == TensorProto.FLOAT16:
            io.type.tensor_type.elem_type = TensorProto.FLOAT
    return graph


def canonicalize_resize(graph: OnnxGraph, op_name: str):
    """The scales input of Resize must be float32."""
    assert op_name in graph.nodes
    node_pb = graph.nodes[op_name]["pb"]
    scales = node_pb.input[2]
    if scales:
        graph.add_onnx_node(cast_in(node_pb, 2, TensorProto.FLOAT))


def dtype_canonicalize(graph: OnnxGraph, op_type: str, op_name: str):
    """Canonicalize data type for operators that don't support fp16."""
    if op_type == "Resize":
        canonicalize_resize(graph, op_name)
    # TODO: add more operators here


@PASSES.register()
def float_to_half(graph: OnnxGraph) -> OnnxGraph:
    """Convert float32 consts and values to half."""

    for node in graph:
        node_pb = graph.nodes[node]["pb"]
        if node_pb.op_type == "Constant":
            tensor = node_pb.attribute[0].t
            if tensor.data_type == TensorProto.FLOAT:
                array = to_array(tensor).astype("float16")
                attr = make_attribute(key="value", value=from_array(array))
                node_pb.attribute.pop()
                node_pb.attribute.append(attr)
    for init in graph.initializer:
        if init.data_type == TensorProto.FLOAT:
            array = to_array(init).astype("float16")
            init.data_type = TensorProto.FLOAT16
            init.raw_data = from_array(array).raw_data
    for io in itertools.chain(graph.input, graph.output):
        if io.type.tensor_type.elem_type == TensorProto.FLOAT:
            io.type.tensor_type.elem_type = TensorProto.FLOAT16
    # fix operators that don't support fp16
    try:
        onnx.checker.check_model(graph.model, full_check=True)
    except onnx.shape_inference.InferenceError as e:
        op_type = re.findall(r"op_type\:(\w+)", f"{e}")[0]
        op_name = re.findall(r"node name\: (.*)\)", f"{e}")[0]
        logger.debug(f"Operator {op_name}({op_type}) check failed, fixing it...")
        dtype_canonicalize(graph, op_type, op_name)
    return graph



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import argparse
import json
from collections import defaultdict
from pathlib import Path
from typing import Optional, Sequence

from . import ONNXIFIER_OPSET, PassManager, convert_graph
from .checker import check_accuracy
from .logger import set_level

USAGE = "onnxify input_model.onnx [output_model.onnx]"


def parse_args():
    parser = argparse.ArgumentParser(
        prog="onnxify",
        usage=USAGE,
        description="onnxify command-line api",
    )
    parser.add_argument(
        "-a",
        "--activate",
        nargs="*",
        help="select passes to be activated, activate L1, L2 and L3 passes if not set.",
    )
    parser.add_argument(
        "-r",
        "--remove",
        nargs="*",
        help="specify passes to be removed from activated passes.",
    )
    parser.add_argument(
        "-n",
        "--no-passes",
        action="store_true",
        help="do not run any optimizing passes, just convert the model",
    )
    parser.add_argument(
        "--print",
        nargs="?",
        const="all",
        help="print the name of all optimizing passes",
    )
    parser.add_argument(
        "--format",
        choices=("protobuf", "textproto", "json", "onnxtxt"),
        default=None,
        help="onnx file format",
    )
    parser.add_argument(
        "-s",
        "--infer-shapes",
        action="store_true",
        help="infer model shapes",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        help="specify a json-format config file for passes",
    )
    parser.add_argument(
        "-u",
        "--uncheck",
        action="store_false",
        help="no checking output model",
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help="check optimized model with random inputs",
    )
    parser.add_argument(
        "--checker-backend",
        choices=("onnx", "openvino", "onnxruntime"),
        default="onnxruntime",
        help="backend for accuracy checking, defaults to openvino",
    )
    parser.add_argument(
        "-v",
        "--opset-version",
        type=int,
        help=f"target opset version, defaults to {ONNXIFIER_OPSET.version}",
    )
    parser.add_argument(
        "-vv",
        "--log-level",
        nargs="?",
        const="DEBUG",
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        help="specify the level of log messages to be printed, defaults to INFO",
    )
    parser.add_argument(
        "-R",
        "--recursive",
        action="store_true",
        help="recursively optimize nested functions",
    )
    parser.add_argument(
        "--nodes",
        nargs="*",
        help="specify a set of node names to apply passes only on these nodes",
    )

    return parser.parse_known_args()


def parse_unknown_args(args: Sequence[str]) -> tuple[str, str, dict, list[str]]:
    """Parse unknown args as pass inputs.

    input_model [output_name] [--key value] ...

    Return:
        str: input model name
        str: output model name
        dict: pass inputs
        list[str]: invalid args
    """

    input_name = ""
    output_name = ""
    configs: dict = {}
    invalid_args: list[str] = []
    value = None
    for i, arg in enumerate(args):
        if arg.startswith("--"):
            if "=" not in arg:
                key = arg[2:]
                if i + 1 < len(args) and not args[i + 1].startswith("--"):
                    value = args[i + 1]
                else:
                    configs[key] = True
                    continue
            else:
                key, value = arg[2:].split("=", 1)
            if key in configs:
                if isinstance(configs[key], list):
                    configs[key].append(value)
                else:
                    configs[key] = [configs[key], value]
            else:
                configs[key] = value
        elif value is not None:
            value = None
            continue
        elif input_name == "":
            input_name = arg
        elif output_name == "":
            output_name = arg
        else:
            invalid_args.append(arg)
    domain_configs: dict[str, dict] = defaultdict(dict)
    for k, v in configs.items():
        if ":" in k:
            domain, key = k.split(":", 1)
        else:
            domain = ""
            key = k
        if domain not in domain_configs:
            domain_configs[domain] = {}
        try:
            domain_configs[domain][key] = json.loads(v)
        except Exception:  # pylint: disable=broad-except
            print(f'[W] use "{v}" directly as it is not a json string')
            domain_configs[domain][key] = v
    return input_name, output_name, domain_configs, invalid_args


def read_configs_from_args_or_file(args, argv) -> tuple[Path, Path, Optional[dict]]:
    """Read configuration from command-line arguments or a config file.

    Args:
        args: Parsed command-line arguments.
        argv: Remaining command-line arguments.

    Returns:
        Path: The path to the input model.
        Path: The path to the output model.
        Optional[dict]: A dictionary containing the configuration, or None if
            no configuration is found.
    """
    configs = None
    input_url, output_url, configs, errors = parse_unknown_args(argv)
    if not input_url:
        print("Usage: " + USAGE)
        raise RuntimeError("missing input model")
    if errors:
        print("Usage: " + USAGE)
        raise RuntimeError(f"invalid args: {' '.join(errors)}")
    input_model = Path(input_url).expanduser()
    if not output_url:
        output_url = input_model.stem + "_o2o"
    output_model = Path(output_url).expanduser()

    if args.activate and len(args.activate) == 1 and "" in configs:
        configs[args.activate[0]].update(configs.pop(""))

    if args.config_file:
        if configs:
            print("[W] configs from command line will be ignored")
        config_file = Path(args.config_file).expanduser()
        with open(config_file, encoding="utf-8") as file:
            configs = json.load(file)
    return input_model, output_model, configs


def main():
    args, argv = parse_args()
    if args.log_level:
        set_level(args.log_level)

    if args.print:
        match args.print:
            case "all":
                PassManager.print_all()
            case "l1":
                PassManager.print_l1()
            case "l2":
                PassManager.print_l2()
            case "l3":
                PassManager.print_l3()
            case _:
                PassManager.print(args.print)
        exit(0)
    try:
        input_model, output_model, configs = read_configs_from_args_or_file(args, argv)
    except RuntimeError as ex:
        print(f"[E] {ex}")
        exit(1)
    # TODO: need to check configs?

    graph = convert_graph(
        model=input_model,
        passes=[] if args.no_passes else args.activate,
        exclude=args.remove,
        onnx_format=args.format,
        configs=configs,
        target_opset=args.opset_version,
        recursive=args.recursive,
        specify_node_names=args.nodes,
    )
    output_model = graph.save(
        output_model,
        format=args.format,
        infer_shapes=args.infer_shapes,
        check=args.uncheck,
    )
    print(f"model saved to {output_model}")
    if args.check and isinstance(output_model, Path):
        error_maps = check_accuracy(
            input_model,
            output_model,
            backend=args.checker_backend,
        )
        for k, v in error_maps.items():
            print(f"{k}: {v}")


if __name__ == "__main__":
    main()



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.reference import ReferenceEvaluator

from onnxifier import PassManager
from onnxifier.graph import OnnxGraph


def _build_matmul_graph():
    """Build a simple MatMul graph: input (2, 3) @ weight (3, 5) -> output (2, 5)"""
    matmul = make_node("MatMul", ["x", "w"], ["y"], "matmul")
    graph = make_graph(
        [matmul],
        "matmul_graph",
        [make_value_info("x", make_tensor_type_proto(1, [2, 3]))],
        [make_value_info("y", make_tensor_type_proto(1, [2, 5]))],
        [make_tensor("w", 1, [3, 5], np.random.rand(3, 5).astype("float32"))],
    )
    return make_model(graph)


def _build_batch_matmul_graph():
    """Build a batch MatMul graph: input (2, 3, 5) @ weight (2, 5, 7)
    -> output (2, 3, 7)
    """
    matmul = make_node("MatMul", ["x", "w"], ["y"], "batch_matmul")
    graph = make_graph(
        [matmul],
        "batch_matmul_graph",
        [make_value_info("x", make_tensor_type_proto(1, [2, 3, 5]))],
        [make_value_info("y", make_tensor_type_proto(1, [2, 3, 7]))],
        [make_tensor("w", 1, [2, 5, 7], np.random.rand(2, 5, 7).astype("float32"))],
    )
    return make_model(graph)


def test_expand_matmul_basic():
    """Test basic 2D MatMul expansion using PassManager and ReferenceEvaluator"""
    # Build original model
    model = _build_matmul_graph()
    runner1 = ReferenceEvaluator(model)

    # Create test input
    x = np.random.uniform(0, 1, size=[2, 3]).astype(np.float32)
    y1_result = runner1.run(None, {"x": x})
    y1 = y1_result[0] if isinstance(y1_result, list) else list(y1_result.values())[0]

    # Apply expansion pass
    graph = OnnxGraph(model)
    pm = PassManager(["expand_matmul"])
    graph = pm.optimize(graph, strict=True)

    # Test with same input shape (padding is handled internally)
    runner2 = ReferenceEvaluator(graph.model)

    # Use the same input - no need to expand manually
    y2_result = runner2.run(None, {"x": x})
    y2 = y2_result[0] if isinstance(y2_result, list) else list(y2_result.values())[0]

    # Results should be identical
    assert np.allclose(
        y1, y2, rtol=1e-5, atol=1e-6
    ), f"Basic MatMul expansion failed. Max diff: {np.max(np.abs(y1 - y2))}"


def test_expand_matmul_batch():
    """Test batch MatMul expansion - batch dimensions should not be expanded"""
    # Build original batch model
    model = _build_batch_matmul_graph()
    runner1 = ReferenceEvaluator(model)

    # Create test input
    x = np.random.uniform(0, 1, size=[2, 3, 5]).astype(np.float32)
    y1_result = runner1.run(None, {"x": x})
    y1 = y1_result[0] if isinstance(y1_result, list) else list(y1_result.values())[0]

    # Apply expansion pass
    graph = OnnxGraph(model)
    pm = PassManager(["expand_matmul"])
    graph = pm.optimize(graph, strict=True)

    # Test with same input shape (padding is handled internally)
    runner2 = ReferenceEvaluator(graph.model)

    # Use the same input - no need to expand manually
    y2_result = runner2.run(None, {"x": x})
    y2 = y2_result[0] if isinstance(y2_result, list) else list(y2_result.values())[0]

    # Results should be identical
    assert np.allclose(
        y1, y2, rtol=1e-5, atol=1e-6
    ), f"Batch MatMul expansion failed. Max diff: {np.max(np.abs(y1 - y2))}"


def test_expand_matmul_no_expansion_needed():
    """Test that already aligned dimensions don't get expanded"""
    # Create a MatMul with dimensions already aligned to factor=2
    matmul = make_node("MatMul", ["x", "w"], ["y"], "matmul")
    graph = make_graph(
        [matmul],
        "aligned_matmul_graph",
        [make_value_info("x", make_tensor_type_proto(1, [4, 6]))],  # Already aligned
        [make_value_info("y", make_tensor_type_proto(1, [4, 6]))],
        [
            make_tensor(
                "w",
                1,
                [6, 6],
                np.random.rand(6, 6).astype("float32"),  # Already aligned
            )
        ],
    )
    model = make_model(graph)

    # Apply expansion pass
    original_graph = OnnxGraph(model)
    pm = PassManager(["expand_matmul"])
    optimized_graph = pm.optimize(original_graph, strict=True)

    # Graph should remain unchanged (same number of nodes)
    assert len(optimized_graph.nodes) == len(
        original_graph.nodes
    ), "Graph was modified when no expansion was needed"


def test_expand_matmul_numerical_accuracy():
    """Comprehensive numerical accuracy test for different shapes"""
    test_cases = [
        ([3, 5], [5, 7], "2D MatMul"),
        ([2, 3, 5], [2, 5, 7], "3D Batch MatMul"),
        ([5, 13], [13, 11], "2D with odd dimensions"),
    ]

    for input_shape, weight_shape, description in test_cases:
        print(f"\nTesting {description}: {input_shape} @ {weight_shape}")

        # Build model for this test case
        matmul = make_node("MatMul", ["x", "w"], ["y"], "matmul")
        output_shape = input_shape[:-1] + [weight_shape[-1]]

        graph = make_graph(
            [matmul],
            f"test_{description.replace(' ', '_').lower()}",
            [make_value_info("x", make_tensor_type_proto(1, input_shape))],
            [make_value_info("y", make_tensor_type_proto(1, output_shape))],
            [
                make_tensor(
                    "w",
                    1,
                    weight_shape,
                    np.random.rand(*weight_shape).astype("float32"),
                )
            ],
        )
        model = make_model(graph)

        # Run original model
        runner1 = ReferenceEvaluator(model)
        x = np.random.uniform(0, 1, size=input_shape).astype(np.float32)
        y1_result = runner1.run(None, {"x": x})
        if isinstance(y1_result, list):
            y1 = y1_result[0]
        else:
            y1 = list(y1_result.values())[0]

        # Apply expansion pass
        graph_obj = OnnxGraph(model)
        pm = PassManager(["expand_matmul"])
        optimized_graph = pm.optimize(graph_obj, strict=True)

        # Run optimized model with same input (padding handled internally)
        runner2 = ReferenceEvaluator(optimized_graph.model)
        y2_result = runner2.run(None, {"x": x})
        if isinstance(y2_result, list):
            y2 = y2_result[0]
        else:
            y2 = list(y2_result.values())[0]

        # Check numerical accuracy
        max_diff = np.max(np.abs(y1 - y2))
        mean_diff = np.mean(np.abs(y1 - y2))

        assert np.allclose(
            y1, y2, rtol=1e-5, atol=1e-6
        ), f"{description} failed. Max diff: {max_diff}, Mean diff: {mean_diff}"


def test_expand_matmul_factor_2():
    """Test with factor=2 to verify the new default"""
    # Test case specifically designed for factor=2
    input_shape = [3, 5]  # Should expand to [4, 6]
    weight_shape = [5, 7]  # Should expand to [6, 8]

    matmul = make_node("MatMul", ["x", "w"], ["y"], "matmul")
    output_shape = input_shape[:-1] + [weight_shape[-1]]

    graph = make_graph(
        [matmul],
        "test_factor_2",
        [make_value_info("x", make_tensor_type_proto(1, input_shape))],
        [make_value_info("y", make_tensor_type_proto(1, output_shape))],
        [
            make_tensor(
                "w", 1, weight_shape, np.random.rand(*weight_shape).astype("float32")
            )
        ],
    )
    model = make_model(graph)

    # Run original model
    runner1 = ReferenceEvaluator(model)
    x = np.random.uniform(0, 1, size=input_shape).astype(np.float32)
    y1_result = runner1.run(None, {"x": x})
    y1 = y1_result[0] if isinstance(y1_result, list) else list(y1_result.values())[0]

    # Apply expansion pass with default factor=2
    graph_obj = OnnxGraph(model)
    pm = PassManager(["expand_matmul"])
    optimized_graph = pm.optimize(graph_obj, strict=True)

    # Run optimized model
    runner2 = ReferenceEvaluator(optimized_graph.model)
    y2_result = runner2.run(None, {"x": x})
    if isinstance(y2_result, list):
        y2 = y2_result[0]
    else:
        y2 = list(y2_result.values())[0]

    # Check numerical accuracy
    max_diff = np.max(np.abs(y1 - y2))
    mean_diff = np.mean(np.abs(y1 - y2))

    assert np.allclose(
        y1, y2, rtol=1e-5, atol=1e-6
    ), f"Factor=2 test failed. Max diff: {max_diff}, Mean diff: {mean_diff}"

    # Verify that expansion actually happened by checking node count
    original_node_count = len(model.graph.node)
    optimized_node_count = len(optimized_graph.model.graph.node)
    assert (
        optimized_node_count > original_node_count
    ), "Expected more nodes after expansion (Pad, expanded MatMul, Slice)"


if __name__ == "__main__":
    test_expand_matmul_basic()
    test_expand_matmul_batch()
    test_expand_matmul_no_expansion_needed()
    test_expand_matmul_numerical_accuracy()
    test_expand_matmul_factor_2()



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.helper import tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from ...checker import show_difference
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register("canonicalize_negative_scale")
class CanonicalizeNegativeScaleRewriter(Rewriter):
    """Negative scale leads to quantization overflow on some devices,
    for negative scales, we should flip the sign and check whether quantized
    value exceeds the range.

    Before:

        0.16 (x) / -0.00125 (scale) = -128 (q)

    After:

        0.15875 (x) / 0.00125 (scale) = 127 (q)
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("DequantizeLinear"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        if prev_n := self.get_input_node(node, 0):
            if prev_n.op_type in ("QuantizeLinear",):
                return  # skip activation quantization
        weight = self.get_value(node.input[0])
        if weight is None:
            return  # skip non constant weight
        scale = self.get_value(node.input[1])
        zero_point = self.get_value(node.input[2])
        if scale is None or zero_point is None:
            raise ValueError("scale and zero_point must be constant")
        if scale.ndim < 1 or np.all(scale >= 0) or np.any(zero_point != 0):
            return  # skip tensor-wise scaling, asymmetric scaling

        axis = self.get_attribute(node, "axis")
        if axis is None:
            axis = 0
        else:
            assert isinstance(axis, (int, float, str))
            axis = int(axis)
        neg_mask = scale < 0
        logger.debug(f"======={node.name}=======")
        logger.debug(f"neg scales {scale[neg_mask]} on channels {np.nonzero(neg_mask)}")
        shape = np.ones([weight.ndim], np.int64)
        shape[axis] = -1
        dq_weight = (weight.astype(np.float32)) * scale.reshape(shape)
        scale = scale.copy()
        scale[neg_mask] = -scale[neg_mask]

        etype = graph.tensor_type(node.input[0])
        dtype = tensor_dtype_to_np_dtype(etype)
        dinfo = np.iinfo(dtype)
        reweight = np.round(dq_weight / scale.reshape(shape)).clip(dinfo.min, dinfo.max)
        reweight = reweight.astype(dtype)
        weight_mask = weight != reweight
        diff = show_difference(
            np.abs(weight.astype("int32")), np.abs(reweight.astype("int32")), rtol=0.01
        )
        logger.debug(f"reweight: {weight[weight_mask]} -> {reweight[weight_mask]}")
        logger.debug(f"diff: {diff}")
        weight_node = make_constant(f"{node.name}/new_weight", reweight)
        scale_node = make_constant(f"{node.name}/new_scale", scale)
        node.input[0] = weight_node.output[0]
        node.input[1] = scale_node.output[0]
        self += [weight_node, scale_node]



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx import NodeProto, numpy_helper
from onnx.helper import make_node

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="constant_to_constantofshape")
class ConstantToConstantOfShapeRewriter(Rewriter):
    """Rewrite Constant to ConstantOfShape node.

    This pass reduces ONNX model size by converting large constant tensors
    to ConstantOfShape nodes when the number of elements exceeds threshold.
    """

    def __init__(self, threshold: int = 16):
        """Initialize the rewriter.

        Args:
            threshold: Only convert constants with more than this many elements.
        """
        super().__init__(pattern=SingleNodePattern("Constant"))
        self.threshold = threshold

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]

        # Get the constant value
        constant_value = self.get_value(node)
        if constant_value is None:
            return

        # Only convert if the number of elements exceeds threshold
        if constant_value.size <= self.threshold:
            return

        # Take the first element as the scalar value to reduce model size
        scalar_value = constant_value.flat[0]

        # Create shape constant for the shape input
        shape_array = np.array(constant_value.shape, dtype=np.int64)
        shape_node = make_constant(node.name + "/shape", shape_array)

        # Create ConstantOfShape node
        # The value attribute should be a tensor with single element
        value_tensor = numpy_helper.from_array(
            np.array([scalar_value], dtype=constant_value.dtype)
        )
        constantofshape_node = make_node(
            op_type="ConstantOfShape",
            inputs=[shape_node.output[0]],
            outputs=node.output,
            name=node.name + "/constantofshape",
            value=value_tensor,
        )

        # Replace the original constant node
        self -= node
        self += shape_node
        self += constantofshape_node



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from abc import ABCMeta, abstractmethod
from collections.abc import Generator, Sequence, Sized
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Set
from uuid import uuid4

import numpy as np
import onnx
from onnx import FunctionProto, numpy_helper
from onnx.helper import make_attribute, make_function, make_operatorsetid

from ..graph import OnnxGraph
from ..logger import error
from .pattern import Pattern
from .utils import attribute_value, evaluate_on_node


class RewriterRepeat(Enum):
    """Repeat mode of the rewriter. By default, the rewriter will only match the graph
    once. If the rewriter still wants to match the rewritten graph recursively, set to
    ``INFINITE``.
    """

    ONCE = 1
    INFINITE = 9999  # maximum repeat times


class MetaRewriter(ABCMeta):
    """Override the attribute of new rewriter class."""

    def __new__(mcs, name, bases, namespace, /, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        setattr(cls, "__NAME__", name)
        setattr(cls, "__DEPS__", [])
        setattr(cls, "__PATCHES__", [])
        return cls


class Rewriter(metaclass=MetaRewriter):
    """OnnxGraph rewriter to modify the graph.

    Args:
        pattern (Pattern): a pattern to match nodes
        repeat (RewriterRepeat): repeat mode of the rewriter. Default is ONCE.
    """

    __NAME__: str = "Rewriter"
    __DEPS__: List[str] = []
    __PATCHES__: List[str] = []

    def __init__(self, pattern: Pattern, repeat=RewriterRepeat.ONCE):
        assert isinstance(pattern, Pattern)
        self.pattern = pattern
        repeat_value = repeat.value if isinstance(repeat, RewriterRepeat) else repeat
        self.repeat = min(int(repeat_value), RewriterRepeat.INFINITE.value)
        self.node_to_add: Set[onnx.NodeProto] = set()
        self.node_to_remove: Set[onnx.NodeProto] = set()
        self.pre_hooks: Dict[int, Callable[[OnnxGraph], OnnxGraph]] = {}
        self.post_hooks: Dict[int, Callable[[OnnxGraph], OnnxGraph]] = {}
        # record how many patterns have been matched and rewritten
        self.num_rewrites = 0

    @abstractmethod
    def rewrite(self, graph: OnnxGraph, nodes: List[onnx.NodeProto], *args, **kwargs):
        """Implement how to rewrite matched nodes in the graph

        Args:
            graph (OnnxGraph): an onnx graph
            nodes (List[NodeProto]): a list of matched nodes
        """

    def add(self, nodes: Sequence[onnx.NodeProto] | onnx.NodeProto):
        """Append nodes to be added to the graph

        Args:
            nodes: a single node or a list of nodes to add
        """
        if isinstance(nodes, onnx.NodeProto):
            self.node_to_add.add(nodes)
        elif isinstance(nodes, Sequence):
            assert all(isinstance(i, onnx.NodeProto) for i in nodes)
            self.node_to_add.update(nodes)
        else:
            raise TypeError(
                f"Expect to add a node or a list of nodes, but got {type(nodes)}."
            )

    def remove(self, nodes: Sequence[onnx.NodeProto] | onnx.NodeProto):
        """Remove nodes from the graph.

        Args:
            nodes: a single node or a list of nodes to remove
        """
        if isinstance(nodes, onnx.NodeProto):
            self.node_to_remove.add(nodes)
        elif isinstance(nodes, Sequence):
            assert all(isinstance(i, onnx.NodeProto) for i in nodes)
            self.node_to_remove.update(nodes)
        else:
            raise TypeError(
                f"Expect to add a node or a list of nodes, but got {type(nodes)}."
            )

    def match_and_rewrite(
        self,
        graph: OnnxGraph,
        *args,
        names: str | Sequence[str] | None = None,
        _specify_node_names: Set[str] | None = None,
        **kwargs,
    ) -> OnnxGraph:
        """Look up for matched patterns in the graph and rewrite it.

        Args:
            graph (OnnxGraph): onnx graph
            names (Sequence[str], optional): Only optimize nodes with these
                names. Defaults to None. A public interface for users.
            _specify_node_names (set[str], optional): Only optimize nodes with these
                names. Defaults to None. A private interface for PassManager.

        Returns:
            OnnxGraph: rewritten graph
        """
        repeat = self.repeat
        self.num_rewrites = 0
        for hook_fn in self.pre_hooks.values():
            graph = hook_fn(graph)
        _names: Set[str] = set()
        if names:
            _names.update([names] if isinstance(names, str) else names)
        if _specify_node_names:
            _names.update(_specify_node_names)
        while repeat > 0:
            matched_nodes = self.pattern.match(graph, _names)
            if matched_nodes is None:
                return graph  # no nothing
            if not isinstance(matched_nodes, Generator):
                raise RuntimeError(
                    f"{type(self.pattern)} match function is invalid! It should yield "
                    f"matched nodes, but it returns {type(matched_nodes)}"
                )
            # pylint: disable=attribute-defined-outside-init
            self._graph = graph
            i = None
            for i, nodes in enumerate(matched_nodes):
                if isinstance(nodes, onnx.NodeProto):
                    nodes = [nodes]
                if self.node_to_remove:
                    # filter nodes that been removed in previous pass
                    nodes = list(filter(lambda n: n not in self.node_to_remove, nodes))
                rewriter_can_run = len(nodes) != 0
                if isinstance(self.pattern, Sized):
                    # shared match been modified by previous rewriter
                    rewriter_can_run = len(nodes) == len(self.pattern)
                if rewriter_can_run:
                    try:
                        self.rewrite(graph, nodes, *args, **kwargs)
                    except Exception:
                        nodes_name = ",".join(n.name for n in nodes)
                        error(f"Rewrite nodes [{nodes_name}] failed.")
                        raise
                    finally:
                        self.num_rewrites += 1
            if i is None:
                break  # no match found
            for node in self.node_to_add:
                graph.add_onnx_node(node)
            added_node_names = set([n.name for n in self.node_to_add])
            for node in self.node_to_remove:
                if node.name in added_node_names:
                    # this node has been replaced by a new node, skip it
                    continue
                graph.remove_onnx_node(node)
            self.node_to_add.clear()
            self.node_to_remove.clear()
            repeat -= 1
        for hook_fn in self.post_hooks.values():
            graph = hook_fn(graph)
        return graph

    def get_input_node(
        self, node: onnx.NodeProto, i_or_s: int | str
    ) -> onnx.NodeProto | None:
        """Get the input node to the i-th input."""
        graph = self.graph
        # pylint: disable=protected-access
        if isinstance(i_or_s, int):
            i = i_or_s
            if i < 0:
                i += len(node.input)
            assert i < len(node.input), f"index {i} exceeds input number"
            if name := graph._out_to_node.get(node.input[i]):
                return graph.nodes[name]["pb"]
        else:
            name = i_or_s
            if name := graph._out_to_node.get(name):
                return graph.nodes[name]["pb"]

    def get_input_node_or_die(
        self, node: onnx.NodeProto, i_or_s: int | str
    ) -> onnx.NodeProto:
        """Get the input node to the i-th input or raise an exception if not found."""
        input_node = self.get_input_node(node, i_or_s)
        if input_node is None:
            raise ValueError(f"Failed to find input node for {node.name}({i_or_s})")
        return input_node

    def get_input_nodes(self, node: onnx.NodeProto) -> List[onnx.NodeProto | None]:
        """Get all input nodes for the given node."""
        return [self.get_input_node(node, i) for i in node.input]

    def get_output_node(
        self, node: onnx.NodeProto, i_or_s: int | str = 0
    ) -> List[onnx.NodeProto]:
        """Get the output node from the i-th output."""
        graph = self.graph
        if isinstance(i_or_s, int):
            i = i_or_s
            if i < 0:
                i += len(node.output)
            assert i < len(node.output), f"index {i} exceeds output number"
            port = node.output[i]
        else:
            port = i_or_s
        return [s for s in graph.onnx_successors(node) if port in s.input]

    def get_attribute(self, node: onnx.NodeProto, name: str, default=None):
        """Try to get the value of an attribute of the node.

        Args:
            node: a node
            name: name of the attribute
            default: set a default value if attribute is optional and not found
        """
        for attr in node.attribute:
            if attr.name == name:
                return attribute_value(attr)
        return default

    def set_attribute(self, node: onnx.NodeProto, name: str, value: Any):
        """Set a new value to an attribute of the node."""
        for attr in node.attribute:
            if attr.name == name:
                match attr.type:
                    case onnx.AttributeProto.FLOAT:
                        attr.f = float(value)
                    case onnx.AttributeProto.INT:
                        attr.i = int(value)
                    case onnx.AttributeProto.STRING:
                        attr.s = str(value).encode()
                    case onnx.AttributeProto.TENSOR:
                        attr.t.CopyFrom(numpy_helper.from_array(value))
                    case onnx.AttributeProto.TYPE_PROTO:
                        attr.tp.CopyFrom(value)
                    case onnx.AttributeProto.FLOATS:
                        attr.ClearField("floats")
                        attr.floats.extend(value)
                    case onnx.AttributeProto.INTS:
                        attr.ClearField("ints")
                        attr.ints.extend(value)
                    case onnx.AttributeProto.STRINGS:
                        attr.ClearField("strings")
                        attr.strings.extend(i.encode() for i in value)
                    case onnx.AttributeProto.TENSORS:
                        attr.ClearField("tensors")
                        attr.tensors.extend(value)
                    case onnx.AttributeProto.GRAPHS:
                        attr.ClearField("graphs")
                        attr.graphs.extend(value)
                    case onnx.AttributeProto.TYPE_PROTOS:
                        attr.ClearField("type_protos")
                        attr.type_protos.extend(value)
                    case _:
                        raise ValueError(f"Unsupported attribute type: {attr.type}")
                return
        node.attribute.append(make_attribute(name, value))

    def remove_attribute(self, node: onnx.NodeProto, name: str):
        """Remove an attribute from the node."""
        for attr in list(node.attribute):
            if attr.name == name:
                node.attribute.remove(attr)
                break

    def get_value(
        self, node: onnx.NodeProto | str, output_name: Optional[str] = None
    ) -> Optional[np.ndarray]:
        """Get value from a constant node."""
        external_base = self.graph.external_base or ""
        if isinstance(node, str):
            for init in self.graph.initializer:
                if init.name == node:
                    return numpy_helper.to_array(init, external_base)
            # value not found in initializer, try to find it from constant.
            # pylint: disable=protected-access
            cst_node_name = self.graph._out_to_node.get(node)
            if not cst_node_name:
                return
            return self.get_value(self.graph.nodes[cst_node_name]["pb"], node)
        elif node.op_type == "Constant":
            attr = node.attribute[0]
            if attr.name == "value":
                return numpy_helper.to_array(attr.t, external_base)
            if attr.name == "value_float":
                return np.array(attr.f, dtype=np.float32)
            if attr.name == "value_floats":
                return np.array(attr.floats, dtype=np.float32)
            if attr.name == "value_int":
                return np.array(attr.i, dtype=np.int64)
            if attr.name == "value_ints":
                return np.array(attr.ints, dtype=np.int64)
            if attr.name == "value_string":
                return np.array(attr.s)
            if attr.name == "value_strings":
                return np.array(attr.strings)
            raise TypeError(f"Unknown attribute type of Constant: {attr.name}")
        elif node.op_type == "Shape":
            return np.asarray(self.graph.tensor_shape(node.input[0]))
        # try to fold all upstreaming nodes
        return evaluate_on_node(self.graph, node, output_name)

    def get_value_or_die(
        self, node: onnx.NodeProto | str, output_name: Optional[str] = None
    ) -> np.ndarray:
        """Get value and raise exception if value is not evaluated."""
        value = self.get_value(node, output_name)
        if value is None:
            node_name = node if isinstance(node, str) else node.name
            raise ValueError(f"Failed to evaluate value for {node_name}({output_name})")
        return value

    def make_function_from_graph(
        self, graph: OnnxGraph, domain: str, type_name: str
    ) -> FunctionProto:
        """Make onnx local function from a subgraph."""
        h = graph.model.graph
        function = make_function(
            domain=domain,
            fname=type_name,
            inputs=list(graph.inputs),
            outputs=list(graph.outputs),
            nodes=h.node,
            opset_imports=[
                make_operatorsetid(domain, 1),
                make_operatorsetid("", graph.opset_version),
            ],
            value_info=h.value_info,
        )
        return function

    def register_post_hook(self, hook: Callable[[OnnxGraph], OnnxGraph]) -> int:
        """Register a post-rewrite hook function.

        A hook function must be a callable that takes an OnnxGraph as input and returns
        an OnnxGraph as output. It will be called after the last run of rewrite.

        Args:
            hook: a hook function to be registered

        Returns:
            int: a unique id of the hook function
        """
        if not callable(hook):
            raise TypeError(f"Expect a callable, but got {type(hook)}")
        uid = uuid4().node
        self.post_hooks[uid] = hook
        return uid

    def remove_post_hook(self, uid: int):
        """Remove a post-rewrite hook function.

        Args:
            uid: a unique id of the hook function to be removed
        """
        self.post_hooks.pop(uid)

    def register_pre_hook(self, hook: Callable[[OnnxGraph], OnnxGraph]) -> int:
        """Register a pre-rewrite hook function.

        A hook function must be a callable that takes an OnnxGraph as input and returns
        an OnnxGraph as output. It will be called before each run of rewrite.

        Args:
            hook: a hook function to be registered

        Returns:
            int: a unique id of the hook function
        """
        if not callable(hook):
            raise TypeError(f"Expect a callable, but got {type(hook)}")
        uid = uuid4().node
        self.pre_hooks[uid] = hook
        return uid

    def remove_pre_hook(self, uid: int):
        """Remove a pre-rewrite hook function.

        Args:
            uid: a unique id of the hook function to be removed
        """
        self.pre_hooks.pop(uid)

    @property
    def graph(self) -> OnnxGraph:
        """Get the current graph."""
        return self._graph

    def __call__(self, graph: OnnxGraph, *args, **kwargs) -> OnnxGraph:
        """Make rewriter a callable."""
        return self.match_and_rewrite(graph, *args, **kwargs)

    def __iadd__(self, nodes: Sequence[onnx.NodeProto] | onnx.NodeProto):
        self.add(nodes)
        return self

    def __isub__(self, nodes: Sequence[onnx.NodeProto] | onnx.NodeProto):
        self.remove(nodes)
        return self



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.onnx_pb import NodeProto

from ...graph import OnnxGraph
from ...logger import warning
from ...passes import PASSES
from ...passes.pattern import SingleNodePattern
from ...passes.rewriter import Rewriter
from ...passes.utils import make_constant


@PASSES.register("pad_remove_axes")
class PadRemoveAxesRewriter(Rewriter):
    """Expand Pad "pads" input and drop the optional "axes" input.

    Note:

        OpenVINO Pad operator does not support "axes" input.

    Before:

        Pad(data, pads, constant_value, axes)

    After:

        Pad(data, expanded_pads, constant_value)
    """

    def __init__(self):
        super().__init__(SingleNodePattern("Pad"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        pad_node = nodes[0]

        if len(pad_node.input) < 4 or not pad_node.input[3]:
            return

        data_input = pad_node.input[0]
        pads_input = pad_node.input[1]
        axes_input = pad_node.input[3]

        try:
            rank = len(graph.tensor_shape(data_input))
        except ValueError as exc:  # tensor shape missing or dynamic rank
            warning(
                f"Pad {pad_node.name} tensor shape unavailable, "
                f"skip removing axes: {exc}"
            )
            return

        pads_value = self.get_value_or_die(pads_input)
        axes_value = self.get_value_or_die(axes_input)

        axes = np.asarray(axes_value).astype(np.int64).tolist()
        if not axes:
            pad_node.input.pop(3)
            return

        axes = [axis + rank if axis < 0 else axis for axis in axes]
        if any(axis < 0 or axis >= rank for axis in axes):
            warning(f"Pad {pad_node.name} has invalid axes {axes}")
            return

        pads_value = np.asarray(pads_value, dtype=np.int64).reshape(-1)
        axis_count = len(axes)
        if pads_value.size != axis_count * 2:
            warning(
                f"Pad {pad_node.name} pads size {pads_value.size} "
                f"mismatches axes length {axis_count}"
            )
            return

        before = pads_value[:axis_count]
        after = pads_value[axis_count:]
        full_pads = np.zeros(rank * 2, dtype=np.int64)
        for idx, axis in enumerate(axes):
            full_pads[axis] = before[idx]
            full_pads[axis + rank] = after[idx]

        pads_const = make_constant(
            f"{pad_node.name}/expanded_pads", full_pads.astype(np.int64)
        )
        pad_node.input[1] = pads_const.output[0]
        pad_node.input.pop(3)

        self += pads_const
